"""
Defines a symbolic representation of a Gaussian job for batch generation
"""

import os, numpy as np, inspect
from collections import OrderedDict, namedtuple
from ..Data import AtomData

__all__ = [
    "GaussianJobTemplater",
    "GaussianJobArray"
]

####################################################################################################################
#
#                                               GaussianJobException
#
class GaussianJobException(Exception):
    pass

####################################################################################################################
#
#                                               GJFOptFormatter
#
class GJFOptFormatter:
    @classmethod
    def format_base_opt(self, opt):
        return (str(opt), True)
    @classmethod
    def format_compound_opt(self, opt):
        return ("({})".format(", ".join(*(self.format_opt(o)[0] for o in opt))), True)
    @classmethod
    def format_bool_opt(self, opt):
        return (opt, False) if opt else ("", "ignore")
    @classmethod
    def format_none_opt(self, opt):
        return ("", "ignore")
    @classmethod
    def format_dict_opt(self, opt):
        chunks=[]
        for k, i in opt.items():
            opt_val, opt_tag = self.format_opt(i)
            if opt_tag is True:
                chunks.append("{}={}".format(k, opt_val))
            elif not isinstance(opt_tag, str) or opt_tag != "ignore":
                chunks.append(opt_val)
        return ("({})".format(", ".join(*chunks)), True)
    @classmethod
    def format_opt(self, opt):
        router = {
            str: self.format_base_opt,
            int: self.format_base_opt,
            tuple: self.format_compound_opt,
            list: self.format_compound_opt,
            bool: self.format_bool_opt,
            dict: self.format_dict_opt,
            OrderedDict: self.format_dict_opt,
            type(None): self.format_none_opt
        }
        meth = router[type(opt)]
        return meth(opt)
    @classmethod
    def format(self, opt_dict, tag):
        chunks=[]
        for k, i in opt_dict.items():
            opt_val, opt_tag = self.format_opt(i)
            if opt_tag is True:
                chunks.append(tag+"{}={}".format(k, opt_val))
            elif not isinstance(opt_tag, str) or opt_tag != "ignore":
                chunks.append(tag+k)
        return "\n".join(chunks)

####################################################################################################################
#
#                                               GaussianJob
#

class GaussianJobTemplater:
    """A class that writes Gaussian .gjf files given a system and config/template options"""

    job_template_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "Templates")

    def __init__(
            self,
            name,
            *args,
            description = None,
            system = None,
            job = None,
            config = None,
            template = "TemplateTerse.gjf",#"Template.gjf",
            footer=None,
            file = None
            ):

        if not isinstance(name, str):
            raise GaussianJobException("{}: name argument must be string instead of {}".format(
                type(self).__name__,
                name
            ))
        self.name = name

        def throw_bad_field(self, field, must, val):
            raise GaussianJobException("{}: '{}' field must be of type {} (got {})".format(
                type(self).__name__, field, must, val
            ))
        if not isinstance(description, str):
            throw_bad_field(self, 'description', str, description)
        self.desc = description
        if not isinstance(system, self.System):
            throw_bad_field(self, 'system', self.System, system)
        self.system = system
        if not isinstance(job, self.Job):
            throw_bad_field(self, 'job', self.Job, job)
        self.job = job
        if not isinstance(config, self.Config):
            throw_bad_field(self, 'config', self.Config, config)
        self.config = config

        if file is None:
            import tempfile as tf
            f = tf.NamedTemporaryFile()
            dd = os.path.dirname(f.name)
            f.close()
            file = os.path.join(dd, name.replace(" ", "_") + ".gjf")
        if not "Chk" in config:
            config["Chk"] = os.path.splitext(file)[0] + ".chk"
        self.file = file
        self.temp = template

        if footer is None:
            footer = ""
        self.footer = footer

    def format(self):
        """
        Formats the job string

        :return:
        :rtype:
        """
        temp = self.temp
        if not ('{header}' in temp and '{job}' in temp):
            if not os.path.isfile(temp):
                temp = os.path.join(self.job_template_dir, temp)
            with open(temp) as t:
                temp = t.read()
        return temp.format(
            header = self.config.format(),
            description = self.desc,
            job = self.job.format(),
            system = self.system.format(job_type = self.job.job_type),
            footer = inspect.cleandoc(self.footer)
        )

    def write(self, file=None):
        """
        Writes the job to a file

        :param file:
        :type file:
        :return:
        :rtype:
        """
        import tempfile as tf

        if file is None:
            file = self.file
        if file is None:
            tmp = tf.NamedTemporaryFile()
            file = tmp.name + '.gjf'
            tmp.close()

        with open(file, 'w+') as gjf:
            gjf.write(self.format())

        return file

    def start(self, *cmd, binary = 'g09', **kwargs):
        """Starts a Gaussian job

        :param cmd:
        :type cmd:
        :param binary:
        :type binary:
        :param kwargs:
        :type kwargs:
        :return: started process
        :rtype:
        """
        import subprocess

        args = []
        args.extend(cmd)
        if len(args) == 0:
            args.append(binary)

        for k,i in kwargs.items():
            if i is True:
                args.append("-"+k)
            elif i is not None:
                args.append("-{}={:!r}".format(k, i))

        return subprocess.Popen(args)

    def run(self, *args, **kwargs):
        job = self.start(*args, **kwargs)
        job.wait()

    def __str__(self):
        return "{}({},\n\tfile = {} )".format(
            type(self).__name__,
            self.format(),
            self.file if not isinstance(self.file, str) else "'{}'".format(self.file)
        )

    class Job(OrderedDict):
        """Inner class for handling the main job options"""
        job_types = dict(
            {
                "SinglePoint": "SP",
                "PartialOptimization" : "POpt",
                "Optimization": "Opt",
                "Frequency": "Freq",

                "Optimization+Frequency": "POpt Freq",

                "ReactionPath": "IRC",
                "TransitionState": "IRCMax",
                "WavefunctionStability": "Stable",
                "BornOppenheimerDynamics": "BOMD",
                "DensityMatrixDynamics": "AOMD"
            },

            **{k:k for k in ["SP", "POpt", "Opt", "Freq", "POpt Freq", "IRC", "IRCMax", "Scan",
                             "Polar", "ADMP", "BOMD", "Force", "Stable", "Volume"]}

        )
        level_of_theory_keys = ['mp2', 'ccsd', 'b2plypd3', 'b3lyp'] # some level of theory stuff...lots of DFT functionals
        basis_set_keys = ["cc", "aug", "sto"] # common level-of-theory/basis set specs...


        def __init__(self, job_type=None, basis_set=None, **kw):
            for k in kw:
                if job_type is None and k in self.job_types:
                    job_type = k
                if basis_set is None:
                    for b in self.basis_set_keys:
                        if k.startswith(b):
                            basis_set = k
                            break
            if job_type is None:
                job_type = "SinglePoint"
            if basis_set is None:
                basis_set = "MP2/aug-cc-pvdz"
            self.job_type = self.job_types[job_type]

            if self.job_type != job_type and job_type in kw:  # this is what I get for allowing things like ReactionPath -_-
                jtv = kw[job_type]
                del kw[job_type]
                kw[self.job_type] = jtv
            elif self.job_type not in kw:
                kw[self.job_type] = True

            self.basis_set = basis_set
            if basis_set not in kw:
                kw[basis_set] = True

            super().__init__(kw)

        def format(self):
            """Returns a formatted version of the job setup

            :return:
            :rtype:
            """
            return GJFOptFormatter.format(self, "#")

        def __str__(self):
            return self.format()

    class Config(OrderedDict):
        """Inner class for handling the main config options"""
        def __init__(self,
                     *args,
                     Mem=None,
                     NProc=None,
                     Chk=None,
                     **kwargs
                     ):
            """
            Transparent wrapper to super().__init__ that
            just provides the ability to get autocompletion
            """
            if len(args) > 0:
                raise GaussianJobException("{} doesn't support positional args".format(
                    type(self).__name__
                ))
            for k,v in (
                    ('Mem', Mem),
                    ('NProc', NProc),
                    ('Chk', Chk),
            ):
                if v is not None:
                    kwargs[k] = v
            super().__init__(kwargs)

        def format(self):
            """Returns a formatted version of the job setup

            :return:
            :rtype:
            """
            return GJFOptFormatter.format(self, "%")

        def __str__(self):
            return self.format()

    class System:
        def __init__(self,
                     charge = 0,
                     molecule = None,
                     vars = None,
                     bonds = None
                     ):
            self.charge = charge
            self.mol = molecule
            self.bonds = bonds
            self.vars = vars

        def format(self, job_type="SP"):
            return {
                "molecule":self.format_molecule(job_type),
                "charge": self.format_charges(),
                "bonds": self.format_bonds()
            }
        def format_molecule(self, job_type):
            blocks, vars = self.prep_mol(self.mol)

            if job_type == "Scan":
                blocks.extend(self.format_scan_vars(vars))
            elif job_type in {"Opt", "POpt"}:
                blocks.extend(self.format_opt_vars(vars))

            elif job_type == "POpt Freq":
                blocks.extend(self.format_opt_vars(vars))

            else:
                blocks.extend(self.format_const_vars(vars))
            return "\n".join(blocks)

        charge_templates = {
            'integer': '{:<3.0f}',
            'charge_template': '{integer} {integer}'
        }
        def format_charges(self, templates=None):
            """
            Formats charge block

            :return:
            :rtype:
            """

            try:
                c, mult = self.charge
            except TypeError:
                c = self.charge
                mult = 1

            if templates is None:
                templates = {}
            templates = dict(self.charge_templates, **templates)
            charge_template = templates['charge_template'].format(**templates)

            return charge_template.format(c, mult)

        bond_templates = {
            'integer':'{:<5.0f}',
            'bond_template': ' {integer} {integer} {integer}'
        }
        def format_bonds(self, templates=None):
            """
            Formats the bonds held in the job spec
            so that they are appropriate for the bottom of the Gaussian
            job spec

            :param templates:
            :type templates:
            :return:
            :rtype:
            """
            b = self.bonds
            if b is None:
                return ""

            if templates is None:
                templates = {}
            templates = dict(self.bond_templates, **templates)
            bond_template = templates['bond_template'].format(**templates)

            bond_block = [ '' ]*len(b)
            for i,bond in enumerate(b):
                bond = bond if len(bond) > 2 else tuple(bond) + (1,)
                bond_block[i] = bond_template.format(*bond)
            return "\n".join(bond_block)

        opt_variable_templates = {
            'name':'{:>8}',
            'real':'{:>12.8f}',
            'integer':'{:<5.0f}',
            'value_template':'  {name} = {real}',
            'scan_template':'  {name} = {real} s {integer} {real}',
            'constant_template':'  {name} = {real} f'
        }
        @classmethod
        def format_opt_vars(cls, vars, templates=None):
            """
            Formats variable definitions for Gaussian optimization jobs

            :param vars:
            :type vars:
            :return:
            :rtype:
            """
            if templates is None:
                templates = {}
            templates = dict(cls.opt_variable_templates, **templates)
            val_temp = templates['value_template'].format(**templates)
            scan_temp = templates['scan_template'].format(**templates)
            const_temp = templates['constant_template'].format(**templates)

            variables = vars["vars"]
            if len(variables) > 0:
                variables_blocks = [" Variables:"]
            else:
                variables_blocks = []
            for k,c in variables.items():
                if c[1] is None:
                    variables_blocks.append(val_temp.format(k, c[0]))
                else:
                    variables_blocks.append(scan_temp.format(k, *c))

            consts = vars["consts"]
            for k,c in consts.items():
                variables_blocks.append(const_temp.format(k, c))

            return variables_blocks

        scan_variable_templates = {
            'name': '{:>8}',
            'real': '{:>12.8f}',
            'integer': '{:<5.0f}',
            'scan_template': '  {name} = {real} {integer} {real}',
            'constant_template': '  {name} = {real} {integer} {real}'
        }
        @classmethod
        def format_scan_vars(cls, vars, templates=None):
            """
            Formats variable definitions for Gaussian scan jobs

            :param vars:
            :type vars:
            :return:
            :rtype:
            """
            if templates is None:
                templates = {}
            templates = dict(cls.scan_variable_templates, **templates)
            scan_temp = templates['scan_template'].format(**templates)
            const_temp = templates['constant_template'].format(**templates)

            variables = vars["vars"]
            if len(variables) > 0:
                variables_blocks = [" Variables:"]
            else:
                variables_blocks = []
            for k,c in variables.items():
                variables_blocks.append(scan_temp.format(k, *c))

            consts = vars["consts"]
            if len(consts) > 0:
                constants_blocks = []#[" Constants:"]
            else:
                constants_blocks = []
            for k,c in consts.items():
                constants_blocks.append(const_temp.format(k, c, 0, 0))

            return variables_blocks + constants_blocks

        sp_variable_templates = {
            'name': '{:>8}',
            'real': '{:>12.8f}',
            'constant_template': '  {name} = {real}'
        }
        @classmethod
        def format_const_vars(cls, vars, templates=None):
            """
            Formats constant block for Gaussian single point jobs

            :param vars:
            :type vars:
            :return:
            :rtype:
            """

            if templates is None:
                templates = {}
            templates = dict(cls.sp_variable_templates, **templates)
            const_temp = templates['constant_template'].format(**templates)

            variables = vars["vars"]
            consts = vars["consts"]
            if len(consts) + len(variables) > 0:
                constants_blocks = [" Variables:"]
            else:
                constants_blocks = []
            for k,c in variables.items():
                constants_blocks.append(const_temp.format(k, *c))
            for k,c in consts.items():
                constants_blocks.append(const_temp.format(k, c))
            return constants_blocks

        @staticmethod
        def get_coord_type(crds):
            """
            Tries to infer coordinates type

            :param crds:
            :type crds:
            :return:
            :rtype:
            """
            crd_type = None
            try:
                csys = crds.system # try to pull the coordinate system directly off the coordinates
            except AttributeError:
                try:
                    spec, zms = crds # if we can't find it try to get the spec and coordinates for a zmat
                    # if not isinstance(spec[0], str):
                    #     raise ValueError("Oops")
                except ValueError:
                    dim0 = len(crds[0]) # if that doesn't work, then we just have a block of coordinates to work with
                    dimn1 = len(crds[-1]) # in that case we simply check lengths to see what we have
                    if dim0 == 1 or dimn1 == 6:
                        crd_type = "zmat"
                    elif dim0 == 3 and dimn1 == 3:
                        crd_type = "cart"
                else:
                    crd_type = "zmatspec"
            else:
                try:
                    from ..Coordinerds import ZMatrixCoordinateSystem, CartesianCoordinateSystem
                    if isinstance(csys, ZMatrixCoordinateSystem):
                        crd_type = "zmat"
                    elif isinstance(csys, CartesianCoordinateSystem):
                        crd_type = "cart"
                except ImportError:
                    name = crd_type.__name__
                    if "ZMatrix" in name:
                        crd_type = "zmat"
                    elif "Cartesian" in name:
                        crd_type = "cart"

            return crd_type

        def prep_vars(self, *vars):
            """
            Prepares variable specifications so that they
            can be nicely fed into the various formatters

            :param vars:
            :type vars:
            :return:
            :rtype:
            """
            var_map = { # to map variables to coordinates
                "vars"  : OrderedDict(),
                "consts": OrderedDict()
            }
            for var_spec in vars:
                if var_spec is not None:
                    for var in var_spec:
                        if isinstance(var, self.Variable) or len(var) == 4:
                            if not isinstance(var, self.Variable):
                                var = self.Variable(*var)
                            start = var.start
                            num   = var.num
                            step  = var.step
                            if isinstance(num, float):
                                num = np.ceil((num - start)/step)
                            var_map["vars"][var.name] = (start, num, step)
                            try:
                                del var_map["consts"][var.name]
                            except KeyError:
                                pass
                        elif isinstance(var, self.Constant):
                            var_map["consts"][var.name] = var.value
                            try:
                                del var_map["vars"][var.name]
                            except KeyError:
                                pass
                        else:
                            if var[0] not in var_map["consts"]:
                                var_map["consts"][var[0]] = var[1]
                                try:
                                    del var_map['vars'][var[0]]
                                except KeyError:
                                    pass
            return var_map

        def prep_mol(self, mol):
            """
            Prepares the atom for formatting, in particular
            making sure isotopic stuff is clean

            :param mol:
            :type mol:
            :return:
            :rtype:
            """
            try:
                ats = mol["atoms"],
                crds = mol["coords"]
            except TypeError:
                try:
                    ats, crds = mol
                except ValueError:
                    ats = mol.atoms
                    crds = mol.crds

            atoms = ['']*len(ats)
            for i, a in enumerate(ats):
                main = AtomData[a, "PrimaryIsotope"]
                if main:
                    atoms[i] = AtomData[a]["Symbol"]
                else:
                    atoms[i] = "{}(Iso={})".format(AtomData[a]["ElementSymbol"], AtomData[a]["MassNumber"])

            # get molspec blocks based on type of coordinates that were fed in
            crd_type = self.get_coord_type(crds)

            var_list = []
            blocks = [None]*len(ats)
            if crd_type == "zmat":
                if len(crds[0]) > 0:
                    crds = [ [] ] + list(crds)
                for zz, a in zip(enumerate(crds), atoms):
                    subblock = [""]*7 # need a new list per line...
                    i,l = zz
                    subblock[0] = a
                    for j,el in enumerate(l):
                        if (j % 2) == 0:
                            subblock[j+1] = int(el)
                        elif isinstance(el, str):
                            # strings get treated as literals
                            subblock[j+1] = el
                        else:
                            var_type = "r" if j == 1 else "a" if j == 3 else "d" if j == 5 else "??"
                            var = var_type + str(i+1)
                            subblock[j+1] = var
                            var_list.append(self.Constant(var, el))

                    blocks[i] = "{:<20} {:<5} {:<12} {:<5} {:<12} {:<5} {:<12}".format(*subblock)

            elif crd_type == "zmatspec":
                # basically like zmat coords but we loop through two things at once
                spec, crds = crds
                if len(crds[0]) > 0:
                    crds = [ [] ] + list(crds)
                    spec = [ [] ] + list(spec)

                for i, csa in enumerate(zip(crds, spec, atoms)):
                    l, s, a = csa
                    subblock = [""]*7 # need a new list per line...
                    subblock[0] = a
                    for j,e in enumerate(s):
                        if j == i:
                            break
                        subblock[1+2*j] = int(e)
                    for j,el in enumerate(l):
                        if j == i:
                            break
                        if isinstance(el, str):
                            # strings get treated as literals
                            subblock[2+2*j] = el
                        else:
                            var_type = "dist" if j == 0 else "angle" if j == 1 else "dihed" if j == 2 else "??"
                            var = var_type + str(i)
                            subblock[2+2*j] = var
                            var_list.append(self.Constant(var, el))
                    blocks[i] = "{:<20} {:<5} {:<12} {:<5} {:<12} {:<5} {:<12}".format(*subblock)
            elif crd_type == "cart":
                for i, ca in enumerate(zip(crds, atoms)):
                    c, a = ca
                    subblock = [None]*4
                    subblock[0] = a
                    for j,el in enumerate(c):
                        if isinstance(el, str):
                            # strings get treated as literals
                            subblock[1+j] = el
                        else:
                            var_type = "x" if j == 0 else "y" if j == 1 else "z" if j == 2 else "??"
                            var = var_type + str(i+1)
                            subblock[1+j] = var
                            var_list.append(self.Constant(var, el))
                    blocks[i] = "{:<20} 0 {:<12} {:<12} {:<12}".format(*subblock)

            var_map = self.prep_vars(var_list, self.vars)
            return blocks, var_map

        Variable = namedtuple("Variable", ["name", "start", "num", "step"])
        Variable.__new__.__defaults__ = (None,) * 4 # start and name really shouldn't take defaults but w/e
        Constant = namedtuple("Constant", ["name", "value"])

class GaussianJobArray:
    """
    Represents a linked set of Gaussian jobs
    """
    def __init__(self, jobs, link="--Link1--"):
        """
        :param jobs:
        :type jobs: Iterable[GaussianJob]
        :param link: link command (defaults to just `--Link1--`)
        :type link: str
        """
        self.jobs = tuple(jobs)
        self.link_cmd = link

    def format(self):
        """
        Formats a linked Gaussian job

        :return:
        :rtype:
        """
        linker="\n{}\n".format(self.link_cmd)
        return linker.join(
            j.format() for j in self.jobs
        )

    def write(self, file):
        """
        Writes a linked Gaussian job to file

        :param file:
        :type file:
        :return:
        :rtype:
        """

        with open(file, 'w+') as gjf:
            gjf.write(self.format())

        return file


"""
Adapted from the Peeves documentation system but tailored for more interactive usage.
"""

__all__ = []
from .DocsBuilder import *; from .DocsBuilder import __all__ as exposed
__all__ += exposed
from .DocWalker import *; from .DocWalker import __all__ as exposed
__all__ += exposed
from .ExamplesParser import *; from .ExamplesParser import __all__ as exposed
__all__ += exposed


"""
Provides a set of decorators that allow code to be agnostic to whether Numba exists or not
"""
import warnings, typing

__all__ = [
    'njit',
    'jit',
    'type_spec',
    'without_numba',
    'numba_decorator',
    'import_from_numba',
    'objmode',
    'prange'
]

class NumbaState:
    numba_disabled = False

class without_numba:
    def __init__(self):
        self._numba_state = None
    def __enter__(self):
        self._numba_state = NumbaState.numba_disabled
        _numba_disabled = True
    def __exit__(self, exc_type, exc_val, exc_tb):
        NumbaState.numba_disabled = self._numba_state
        self._numba_state = None

def load_numba(warn=False):
    if NumbaState.numba_disabled:
        return None
    try:
        import numba
    except ImportError:
        if isinstance(warn, str) and warn == 'raise':
            raise
        if warn:
            warnings.warn("Numba not installed/code will be slower")
        numba = None

    return numba

def njit(*args, warn=False, **kwargs):
    return numba_decorator(*args, method='njit', warn=warn, **kwargs)

def jit(*args, warn=False, nopython=False, **kwargs):
    return numba_decorator(*args, method='jit', warn=warn, nopython=nopython, **kwargs)

def numba_decorator(*args, method=None, warn=False, **kwargs):
    numba = load_numba(warn=warn)
    if numba is not None:
        return getattr(numba, method)(*args, **kwargs)
    else:
        if len(args) > 0:
            return args[0]
        else:
            return lambda f:f

def type_spec(t, warn=False):
    numba = load_numba(warn=warn)
    if numba is not None:
        return getattr(numba, t)
    else:
        return typing.Any

def import_from_numba(name, default):
    numba = load_numba()
    if numba is not None:
        return getattr(numba, name)
    else:
        return default

class _noop_context:
    def __init__(self, *args, **kwargs):
        pass
    def __enter__(self):
        pass
    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

objmode = import_from_numba('objmode', _noop_context)
prange = import_from_numba('prange', range)


from .CommonData import DataHandler, DataRecord

__all__ = [ "WavefunctionData" ]
__reload_hook__ = [".CommonData"]

class WavefunctionDataHandler(DataHandler):
    def __init__(self):
        super().__init__("WavefunctionData", record_type=WavefunctionDataRecord)

class WavefunctionDataRecord(DataRecord):
    """
    Represents a simple callable wavefunction...
    """
    def __call__(self, *args, **kwargs):
        return self['Wavefunction'](*args, **kwargs)
WavefunctionData=WavefunctionDataHandler()
WavefunctionData.__doc__ = """An instance of WavefunctionDataHandler that can be used for looking up data on wavefunctions"""
WavefunctionData.__name__ = "WavefunctionData"


"""
Provides customized set operations based off of the NumPy builtins
to minimize things like excess sorts
"""

import numpy as np, itertools, math
from .Misc import flatten_dtype, unflatten_dtype, recast_permutation, downcast_index_array, is_numeric

__all__ = [
    'unique',
    'intersection',
    'contained',
    'difference',
    'find',
    'argsort',
    'group_by',
    'grouping_info',
    'take_where_groups',
    'split_by_regions',
    "combination_indices",
    "permutation_indices",
    "vector_ix",
    "vector_take",
    "index_mask",
    "index_complement"
]

coerce_dtype = flatten_dtype
uncoerce_dtype = unflatten_dtype

def argsort(ar):
    ar = np.asanyarray(ar)
    if ar.ndim > 1:
        ar, _, _, _ = coerce_dtype(ar)
    return recast_permutation(np.argsort(ar, kind='mergesort'))

def unique(ar, return_index=False, return_inverse=False,
           return_counts=False, axis=0, sorting=None, minimal_dtype=False):
    """
    A variant on np.unique with default support for `axis=0` and sorting
    """

    ar = np.asanyarray(ar)
    if ar.ndim == 1:
        ret = unique1d(ar, return_index=return_index, return_inverse=return_inverse,
                       return_counts=return_counts, sorting=sorting, minimal_dtype=minimal_dtype)
        return ret

    # axis was specified and not None
    try:
        ar = np.moveaxis(ar, axis, 0)
    except np.AxisError:
        # this removes the "axis1" or "axis2" prefix from the error message
        raise np.AxisError(axis, ar.ndim)

    # coerce the data into the approrpriate shape
    consolidated, dtype, orig_shape, orig_dtype = coerce_dtype(ar)

    output = unique1d(consolidated,
                                return_index=return_index, return_inverse=return_inverse,
                                return_counts=return_counts, sorting=sorting, minimal_dtype=minimal_dtype)
    output = (uncoerce_dtype(output[0], orig_shape, orig_dtype, axis),) + output[1:]
    return output

def unique1d(ar, return_index=False, return_inverse=False,
              return_counts=False, sorting=None, minimal_dtype=False):
    """
    Find the unique elements of an array, ignoring shape.
    """
    ar = np.asanyarray(ar)

    if sorting is None:
        if minimal_dtype:
            sorting = recast_permutation(ar.argsort(kind='mergesort')) # we want to have stable sorts throughout
        else:
            sorting = ar.argsort(kind='mergesort') # we want to have stable sorts throughout
    ar = ar[sorting]

    mask = np.empty(ar.shape, dtype=np.bool_)
    mask[:1] = True
    mask[1:] = ar[1:] != ar[:-1]

    ret = (ar[mask], sorting)
    if return_index:
        ret += (sorting[mask],)
    if return_inverse:
        imask = np.cumsum(mask) - 1
        inv_idx = np.empty(mask.shape, dtype=np.intp)
        inv_idx[sorting] = imask
        ret += (inv_idx,)
    if return_counts:
        idx = np.concatenate(np.nonzero(mask) + ([mask.size],))
        ret += (np.diff(idx),)
    return ret

def intersection(ar1, ar2,
                assume_unique=False, return_indices=False,
                sortings=None, union_sorting=None, minimal_dtype=False
                ):

    ar1 = np.asanyarray(ar1)
    ar2 = np.asanyarray(ar2)

    if ar1.dtype < ar2.dtype:
        ar1 = ar1.astype(ar2.dtype)
    elif ar1.dtype < ar2.dtype:
        ar2 = ar2.astype(ar1.dtype)

    if ar1.ndim == 1:
        ret = intersect1d(ar1, ar2, assume_unique=assume_unique, return_indices=return_indices,
                          sortings=sortings, union_sorting=union_sorting, minimal_dtype=minimal_dtype)
        return ret

    ar1, dtype, orig_shape1, orig_dtype1 = coerce_dtype(ar1)
    ar2, dtype, orig_shape2, orig_dtype2 = coerce_dtype(ar2, dtype=dtype)
    output = intersect1d(ar1, ar2, assume_unique=assume_unique, return_indices=return_indices,
                          sortings=sortings, union_sorting=union_sorting)
    output = (uncoerce_dtype(output[0], orig_shape1, orig_dtype1, None),) + output[1:]
    return output

def intersect1d(ar1, ar2,
                assume_unique=False, return_indices=False,
                sortings=None, union_sorting=None, minimal_dtype=False
                ):
    """
    Find the intersection of two arrays.

    """
    ar1 = np.asanyarray(ar1)
    ar2 = np.asanyarray(ar2)

    if not assume_unique:
        if return_indices:
            if sortings is not None:
                ar1, sorting1, ind1 = unique1d(ar1, return_index=True, sorting=sortings[0], minimal_dtype=minimal_dtype)
                ar2, sorting2, ind2 = unique1d(ar2, return_index=True, sorting=sortings[1], minimal_dtype=minimal_dtype)
            else:
                ar1, sorting1, ind1 = unique1d(ar1, return_index=True, minimal_dtype=minimal_dtype)
                ar2, sorting2, ind2 = unique1d(ar2, return_index=True, minimal_dtype=minimal_dtype)
        else:
            if sortings is not None:
                ar1, sorting1 = unique1d(ar1, sorting=sortings[0], minimal_dtype=minimal_dtype)
                ar2, sorting2 = unique1d(ar2, sorting=sortings[1], minimal_dtype=minimal_dtype)
            else:
                ar1, sorting1 = unique1d(ar1)
                ar2, sorting2 = unique1d(ar2)
        sortings = (sorting1, sorting2)

    aux = np.concatenate((ar1, ar2))
    if union_sorting is None:
        aux_sort_indices = np.argsort(aux, kind='mergesort')
        if minimal_dtype:
            aux_sort_indices = recast_permutation(aux_sort_indices)
        aux = aux[aux_sort_indices]
    else:
        aux_sort_indices = union_sorting
        aux = aux[aux_sort_indices]

    mask = aux[1:] == aux[:-1]
    int1d = aux[:-1][mask]

    if return_indices:
        ar1_indices = aux_sort_indices[:-1][mask]
        ar2_indices = aux_sort_indices[1:][mask] - ar1.size
        if not assume_unique:
            ar1_indices = ind1[ar1_indices]
            ar2_indices = ind2[ar2_indices]
        return int1d, sortings, union_sorting, ar1_indices, ar2_indices
    else:
        return int1d, sortings, union_sorting

def contained(ar1, ar2, assume_unique=False, invert=False,
                sortings=None, union_sorting=None, method=None):
    """
    Test whether each element of `ar1` is also present in `ar2`.
    """

    # Ravel both arrays, behavior for the first array could be different
    ar1 = np.asanyarray(ar1)
    ar2 = np.asanyarray(ar2)

    if ar1.dtype < ar2.dtype:
        ar1 = ar1.astype(ar2.dtype)
    elif ar2.dtype < ar1.dtype:
        ar2 = ar2.astype(ar1.dtype)

    if ar1.ndim > 1:
        ar1, dtype, orig_shape1, orig_dtype1 = coerce_dtype(ar1)
        ar2, dtype, orig_shape2, orig_dtype2 = coerce_dtype(ar2, dtype=dtype)

    # Check if one of the arrays may contain arbitrary objects
    contains_object = ar1.dtype.hasobject or ar2.dtype.hasobject

    # This code is run when
    # a) the first condition is true, making the code significantly faster
    # b) the second condition is true (i.e. `ar1` or `ar2` may contain
    #    arbitrary objects), since then sorting is not guaranteed to work
    if len(ar2) < 10 * len(ar1) ** 0.145 or contains_object:
        if invert:
            mask = np.ones(len(ar1), dtype=bool)
            for a in ar2:
                mask &= (ar1 != a)
        else:
            mask = np.zeros(len(ar1), dtype=bool)
            for a in ar2:
                mask |= (ar1 == a)
        return mask, sortings, union_sorting

    # Otherwise use sorting
    if assume_unique is not True and assume_unique is not False: # i.e. it's a bool
        assume_unique_1, assume_unique_2 = assume_unique
    else:
        assume_unique_1 = assume_unique_2 = assume_unique
    if not assume_unique_1:
        if sortings is None:
            ar1, sorting1, rev_idx = unique1d(ar1, return_inverse=True)
        else:
            ar1, sorting1, rev_idx = unique1d(ar1, sorting=sortings[0], return_inverse=True)
    else:
        if sortings is not None:
            sorting1 = sortings[0]
        else:
            sorting1 = None

    if not assume_unique_2:
        if sortings is None:
            ar2, sorting2 = unique1d(ar2)
        else:
            ar2, sorting2 = unique1d(ar2, sorting=sortings[1])
    else:
        if sortings is not None:
            sorting2 = sortings[1]
        else:
            sorting2 = None
    sortings = (sorting1, sorting2)

    if method is not None and method == 'find':
        find_pos, _ = find(ar2, ar1, sorting='sorted', check=False) #binary search is fast
        if invert:
            ret = ar2[find_pos] != ar1
        else:
            ret = ar2[find_pos] == ar1
        order = None
    else:
        ar = np.concatenate((ar1, ar2))
        # We need this to be a stable sort, so always use 'mergesort'
        # here. The values from the first array should always come before
        # the values from the second array.
        if union_sorting is None:
            order = ar.argsort(kind='mergesort')
        else:
            order = union_sorting
        sar = ar[order]
        if invert:
            bool_ar = (sar[1:] != sar[:-1])
        else:
            bool_ar = (sar[1:] == sar[:-1])
        flag = np.concatenate((bool_ar, [invert]))
        ret = np.empty(ar.shape, dtype=bool)
        ret[order] = flag

    if assume_unique_1:
        return ret[:len(ar1)], sortings, order
    else:
        return ret[rev_idx], sortings, order

def difference(ar1, ar2, assume_unique=False, sortings=None, method=None, union_sorting=None):
    """
    Calculates set differences over any shape of array
    """

    ar1 = np.asanyarray(ar1)
    ar2 = np.asanyarray(ar2)

    if ar1.dtype < ar2.dtype:
        ar1 = ar1.astype(ar2.dtype)
    elif ar2.dtype < ar1.dtype:
        ar2 = ar2.astype(ar1.dtype)

    if ar1.ndim == 1:
        ret = difference1d(ar1, ar2, assume_unique=assume_unique, method=method,
                          sortings=sortings, union_sorting=union_sorting)
        return ret

    ar1, dtype, orig_shape1, orig_dtype1 = coerce_dtype(ar1)
    ar2, dtype, orig_shape2, orig_dtype2 = coerce_dtype(ar2, dtype=dtype)
    output = difference1d(ar1, ar2, assume_unique=assume_unique, method=method,
                          sortings=sortings, union_sorting=union_sorting)
    output = (uncoerce_dtype(output[0], orig_shape1, orig_dtype1, None),) + output[1:]
    return output

def difference1d(ar1, ar2, assume_unique=False, sortings=None, method=None, union_sorting=None):
    """
    Calculates set differences in 1D
    """

    if not assume_unique:
        if sortings is not None:
            ar1, sorting1 = unique(ar1, sorting=sortings[0])
            ar2, sorting2 = unique(ar2, sorting=sortings[0])
        else:
            ar1, sorting1 = unique(ar1)
            ar2, sorting2 = unique(ar2)
        sortings = (sorting1, sorting2)

    in_spec = contained(ar1, ar2, sortings=sortings, union_sorting=union_sorting, assume_unique=True, method=method, invert=True)
    return (ar1[in_spec[0]],) + in_spec[1:]

def find1d(ar, to_find, sorting=None,
           search_space_sorting=None, return_search_space_sorting=False,
           check=True, minimal_dtype=False, missing_val='raise'
           ):
    """
    Finds elements in an array and returns sorting
    """

    presorted = isinstance(sorting, str) and sorting == 'sorted'
    if sorting is None:
        sorting = np.argsort(ar, kind='mergesort')

    if search_space_sorting is None and return_search_space_sorting:
        search_space_sorting = np.argsort(to_find, kind='mergesort')

    if search_space_sorting is not None:
        if isinstance(search_space_sorting, np.ndarray):
            search_space_inverse_sorting = np.argsort(search_space_sorting)
        else:
            search_space_sorting, search_space_inverse_sorting = search_space_sorting

        to_find = to_find[search_space_sorting]

    if presorted:
        vals = np.searchsorted(ar, to_find)
    else:
        vals = np.searchsorted(ar, to_find, sorter=sorting)
    if isinstance(vals, (np.integer, int)):
        vals = np.array([vals])
    # we have the ordering according to the _sorted_ version of `ar`
    # so now we need to invert that back to the unsorted version
    if len(sorting) > 0:
        big_vals = vals == len(ar)
        if np.any(big_vals):
            vals = vals.astype(int)
            vals[big_vals] = -1
        if not presorted:
            vals = sorting[vals]
        if check:
            # now because of how searchsorted works, we need to check if the found values
            # truly agree with what we asked for
            bad_vals = ar[vals] != to_find
            if vals.shape == ():
                if bad_vals:
                    vals = -1
            else:
                # print(vals, bad_vals)
                vals = vals.astype(int)
                vals[bad_vals] = -1
        else:
            bad_vals = big_vals

    else:
        bad_vals = np.full_like(to_find, True)
        vals = np.full_like(vals, -1)
    if check and bad_vals.any():
        if isinstance(missing_val, str) and missing_val == 'raise':
            raise IndexError("{} not in array".format(to_find[bad_vals]))
        else:
            vals = vals.astype(int)
            vals[bad_vals] = missing_val

    if minimal_dtype and not bad_vals.any(): # protecting the missings
        vals = downcast_index_array(vals, ar.shape[-1])

    if search_space_sorting is not None:
        vals = vals[search_space_inverse_sorting]
    ret = (vals, sorting,)
    if return_search_space_sorting:
        ret += ((search_space_sorting, search_space_inverse_sorting),)
    return ret

def find(ar, to_find, sorting=None,
         search_space_sorting=None,
         return_search_space_sorting=False,
         check=True, minimal_dtype=False, missing_val='raise'):
    """
    Finds elements in an array and returns sorting
    """

    ar = np.asanyarray(ar)
    to_find = np.asanyarray(to_find)

    if ar.dtype < to_find.dtype:
        ar = ar.astype(to_find.dtype)
    elif to_find.dtype < ar.dtype:
        to_find = to_find.astype(ar.dtype)

    # print(ar.dtype, to_find.dtype )

    if ar.ndim == 1:
        ret = find1d(ar, to_find, sorting=sorting, check=check,
                     search_space_sorting=search_space_sorting,
                     return_search_space_sorting=return_search_space_sorting,
                     minimal_dtype=minimal_dtype, missing_val=missing_val
                     )
        return ret

    ar, dtype, orig_shape1, orig_dtype1 = coerce_dtype(ar)
    to_find, dtype, orig_shape2, orig_dtype2 = coerce_dtype(to_find, dtype=dtype)
    output = find1d(ar, to_find, sorting=sorting, check=check,
                     search_space_sorting=search_space_sorting,
                     return_search_space_sorting=return_search_space_sorting,
                     minimal_dtype=minimal_dtype, missing_val=missing_val
                    )
    return output

def group_by_spec1d(keys, sorting=None, return_sizes=False):
    uinds, sorting, mask = unique(keys, sorting=sorting, return_inverse=True)
    _, _, inds = unique(mask[sorting], sorting=np.arange(len(mask)), return_index=True)

    ret = (uinds, inds, sorting)
    if return_sizes:
        sizes = np.concatenate([[inds[0]], np.diff(inds)], axis=0)
        ret += (sizes,)
    return ret

def group_by1d(ar, keys, sorting=None, return_sizes=False, return_indices=False):
    """
    Splits an array by a keys
    :param ar:
    :type ar:
    :param keys:
    :type keys:
    :param sorting:
    :type sorting:
    :return:
    :rtype:
    """

    spec = group_by_spec1d(keys, sorting=sorting, return_sizes=return_sizes)
    if return_sizes:
        sizes = spec[-1]
        spec = spec[:-1]
    else:
        sizes = None
    uinds, inds, sorting = spec
    groups = np.split(ar[sorting,], inds)[1:]

    ret = ((uinds, groups), sorting)
    if return_indices:
        ret += (inds,)
    if return_sizes:
        ret += (sizes,)
    return ret

def grouping_info(keys, sorting=None, return_sizes=False):
    """
    Grouping info for keys

    :param keys:
    :type keys:
    :param sorting:
    :type sorting:
    :return: group pairs & sorting info
    :rtype:
    """

    keys = np.asanyarray(keys)
    if keys.ndim == 1:
        ret = group_by_spec1d(keys, sorting=sorting, return_sizes=return_sizes)
        return ret

    keys, dtype, orig_shape, orig_dtype = coerce_dtype(keys)
    output = group_by_spec1d(keys, sorting=sorting, return_sizes=return_sizes)
    ukeys = output[0]
    ukeys = uncoerce_dtype(ukeys, orig_shape, orig_dtype, None)
    output = (ukeys,) + output[1:]
    return output

def group_by(ar, keys, sorting=None, return_sizes=False, return_indices=False):
    """
    Groups an array by keys

    :param ar:
    :type ar:
    :param keys:
    :type keys:
    :param sorting:
    :type sorting:
    :return: group pairs & sorting info
    :rtype:
    """

    ar = np.asanyarray(ar)
    keys = np.asanyarray(keys)

    if keys.ndim == 1:
        ret = group_by1d(ar, keys, sorting=sorting, return_sizes=return_sizes, return_indices=return_indices)
        return ret

    keys, dtype, orig_shape, orig_dtype = coerce_dtype(keys)
    output = group_by1d(ar, keys, sorting=sorting, return_sizes=return_sizes, return_indices=return_indices)
    ukeys, groups = output[0]
    ukeys = uncoerce_dtype(ukeys, orig_shape, orig_dtype, None)
    output = ((ukeys, groups),) + output[1:]
    return output

def split_by_regions1d(ar, regions, sortings=None, return_indices=False):
    """
    :param regions:
    :type regions:
    :param ar1:
    :type ar1:
    :return:
    :rtype:
    """

    if sortings is None:
        sortings = (None, None)
    ar_sorting, region_sorting = sortings
    if ar_sorting is None:
        ar_sorting = argsort(ar)
    ar = ar[ar_sorting]
    if region_sorting is None:
        region_sorting = argsort(regions)

    insertion_spots = np.searchsorted(regions, ar, sorter=region_sorting)
    uinds, _, inds = unique(insertion_spots, sorting=np.arange(len(insertion_spots)), return_index=True)
    groups = np.split(ar, inds)[1:]

    output = (uinds, groups)
    if return_indices:
        return output, inds, sortings
    else:
        return output, sortings

def split_by_regions(ar, regions, sortings=None, return_indices=False):
    """
    Splits an array up by edges defined by regions.
    Operates in 1D but can take compound dtypes using lexicographic
    ordering.
    In that case it is on the user to ensure that lex ordering is what is desired.
    :param ar:
    :type ar:
    :param regions:
    :type regions:
    :param sortings:
    :type sortings:
    :return:
    :rtype:
    """

    ar = np.asanyarray(ar)
    regions = np.asanyarray(regions)

    if ar.ndim == 1:
        ret = split_by_regions1d(regions, ar, sortings=sortings, return_indices=return_indices)
        return ret

    ar, dtype, orig_shape, orig_dtype = coerce_dtype(ar)
    regions, dtype, orig_shape1, orig_dtype1 = coerce_dtype(ar, dtype=dtype)
    output = split_by_regions1d(regions, ar, sortings=sortings, return_indices=return_indices)
    uinds, groups = output[0]
    groups = uncoerce_dtype(groups, orig_shape, orig_dtype, None)
    output = ((uinds, groups),) + output[1:]

    return output

class version_info:
    numpy_version = None
    @classmethod
    def get_np_version(cls):
        if cls.numpy_version is None:
            cls.numpy_version = tuple(int(x) for x in np.version.full_version.split("."))
        return cls.numpy_version
def from_iter_nd(iter, dtype, shape, like=None, **extra):
    version = version_info.get_np_version()
    if version[0] > 1 or version[1] > 22:
        return np.fromiter(
            iter,
            count=shape[0],
            dtype=np.dtype((dtype, shape[1:])),
            **extra
        )
    else:
        if like is None:
            base_array = np.empty(shape, dtype)
        else:
            # probably wrong, but here's the idea
            base_array = like(shape, dtype)
        for i,elem in enumerate(iter):
            base_array[i] = elem
        return base_array
def permutation_indices(n, r, dtype=int):
    if r == 0:
        return np.array([[]])
    return from_iter_nd(
        itertools.permutations(range(n), r),
        shape=(math.factorial(n)//math.factorial(n-r), r),
        dtype=dtype
    )
def combination_indices(n, r, dtype=int):
    if r == 0:
        return np.array([[]])
    elif r == n:
        return np.arange(n)[np.newaxis]
    return from_iter_nd(
        itertools.combinations(range(n), r),
        shape=(math.comb(n, r), r),
        dtype=dtype
    )

def vector_ix(shape, inds, return_shape=False):
    if isinstance(inds, tuple):
        if is_numeric(inds[0]): inds = tuple(np.asanyarray(ii) for ii in inds)
    else:
        inds = (np.asanyarray(inds),)
    if is_numeric(shape, int): shape = (shape,)
    if len(inds) != len(shape): raise ValueError("shape mismatch")
    extra = inds[0].shape[:-1]
    ee = len(extra)
    xx = list(range(ee))
    mask_inds = tuple(
        np.expand_dims(np.arange(s), xx[:i] + xx[i + 1:] + [ee])
        for i, s in enumerate(extra)
    ) + inds
    if return_shape:
        return mask_inds, extra + shape
    else:
        return mask_inds

def index_mask(shape, inds, complement=False):
    inds, shape = vector_ix(shape, inds, return_shape=True)
    mask = np.full(shape, True if complement else False)
    mask[inds] = False if complement else True
    return mask

def index_complement(shape, inds):
    if isinstance(inds, tuple):
        if is_numeric(inds[0]): inds = tuple(np.asanyarray(ii) for ii in inds)
    else:
        inds = (np.asanyarray(inds),)
    if is_numeric(shape): shape = (shape,)
    extra = inds[0].shape[:-len(shape)]
    mask = index_mask(shape, inds, complement=True)
    mask_inds = np.where(mask)
    comp_shape = tuple(s - i for s,i in zip(shape, inds[0].shape[-len(shape):]))
    targ_shape = extra + comp_shape
    return tuple(
        mask_inds[k].reshape(targ_shape)
        for k in range(-len(shape), 0)
    )

def vector_take(arr, inds, shared=None, return_spec=False):
    """
    A generalized array indexing that broadcasts properly across everything except for the specified "take" index
    :param arr:
    :param inds:
    :return:
    """
    arr = np.asanyarray(arr)
    if shared is None and not return_spec: # so common we have to...
        inds = np.asanyarray(inds)
        flat_inds = inds.flatten()
        flat_vals = np.reshape(arr[..., flat_inds], arr.shape[:-1] + inds.shape)
        return flat_vals

    if isinstance(inds, tuple):
        if is_numeric(inds[0]): inds = tuple(np.asanyarray(ii) for ii in inds)
    else:
        inds = (np.asanyarray(inds),)

    if shared is None: shared = 0
    shared_shape = arr.shape[:shared]
    bcast_shape_ind = inds[-1].shape[shared:-1]
    bcast_shape_arr = arr.shape[shared:-len(inds)]

    total_shape = shared_shape + bcast_shape_arr + bcast_shape_ind

    arr_pad_dims = shared + len(bcast_shape_arr)
    arr = np.broadcast_to(
        np.expand_dims(arr, list(range(arr_pad_dims, arr_pad_dims+len(bcast_shape_ind)))),
        total_shape + arr.shape[-len(inds):]
    )

    inds_nob = tuple(
        np.broadcast_to(
            np.expand_dims(x, list(range(shared, shared + len(bcast_shape_arr)))),
            total_shape + x.shape[-len(inds):]
        )
        for x in inds
    )
    inds = vector_ix(arr.shape[-len(inds):], inds_nob)
    if return_spec:
        return arr, inds
    else:
        return arr[inds]

def take_where_groups(arr, where, presorted=True, return_rows=False):
    if len(where) == 1:
        raise ValueError("can't group 1D data")
    vals = arr[where]
    if len(where) > 2:
        keys = np.array(where[:-1]).T
    else:
        keys = where[0]
    if presorted:
        sorting = np.arange(len(keys))
    else:
        sorting = None
    rows, splits, sorting, sizes = grouping_info(where[0], sorting=sorting, return_sizes=True)
    if not presorted:
        presorted = np.equal(np.arange(len(keys)), sorting).all()
    if presorted and len(np.unique(sizes)) == 1:
        # we can just reshape vals
        vals = vals.reshape(-1, sizes[0])
    else:
        vals = np.split(vals[sorting,], splits)[1:]

    if return_rows:
        return (rows, vals)
    else:
        return vals




"""
Simple utilities that support constructing Regex patterns
"""

import re
from collections import OrderedDict
from .StructuredType import StructuredType, DisappearingType
__reload_hook__ = [".StructuredType"]

__all__ = [
    "RegexPattern",
    "Capturing",
    "NonCapturing",
    "Optional",
    "Alternatives",
    "Longest",
    "Shortest",
    "Repeating",
    "Duplicated",
    "PatternClass",
    "Parenthesized",
    "Named",
    "Any",
    "Sign",
    "Number",
    "Integer",
    "PositiveInteger",
    "ASCIILetter",
    "AtomName",
    "WhitespaceCharacter",
    "Whitespace",
    "Word",
    "WordCharacter",
    "VariableName",
    "CartesianPoint",
    "IntXYZLine",
    "XYZLine",
    "Empty",
    "Newline",
    "ZMatPattern"
]

######################################################################################################################
#
#                                            RegexPattern
#
class RegexPattern:
    """
    Represents a combinator structure for building more complex regexes

    It might be worth working with this combinator structure in a _lazy_ fashion so that we can drill down
    into the expression structure... that way we can define a sort-of Regex calculus that we can use to build up higher
    order regexes but still be able to recursively inspect subparts?
    """
    def __init__(self,
                 pat,
                 name = None,
                 children = None,
                 parents = None,
                 dtype = None,
                 repetitions = None,
                 key = None,
                 joiner = "",
                 join_function = None,
                 wrapper_function = None,
                 suffix = None, # often we need a start or end RegexPattern for the match, but we don't actually care the data in it
                 prefix = None,
                 parser = None, # a parser function we can use instead of the default StringParser parser
                 handler = None, # a data handler for after capturing a block
                 default_value = None, # default value for Optional types
                 capturing = None,
                 allow_inner_captures = False # whether or not to multiple captures for a given pattern
                 ):
        """
        :param pat:
        :type pat: str | callable
        :param name:
        :type name: str
        :param dtype:
        :type dtype:
        :param repetitions:
        :type repetitions:
        :param key:
        :type key:
        :param joiner:
        :type joiner:
        :param children:
        :type children:
        :param parents:
        :type parents:
        :param wrapper_function:
        :type wrapper_function:
        :param suffix:
        :type suffix:
        :param prefix:
        :type prefix:
        :param parser:
        :type parser:
        :param handler:
        :type handler:
        :param capturing:
        :type capturing:
        :param allow_inner_captures:
        :type allow_inner_captures:
        """

        if isinstance(pat, (tuple, list)) and children is None:
            children = pat
            pat = lambda p, no_capture = False:p # basically represents a compound pattern...
        self._pat = pat
        self._cached = None
        self._comp = None

        self.name = name # not used for anything but can make the code clearer

        # will some day be useful in automatic type determination in a StringParser
        self._dtype = dtype
        self.repetitions = repetitions

        self.key = key # to be used when matching to find subcomponents

        # we'll build a regex tree structure
        # that we can use to walk the regex string structure
        if children is None:
            children = []
        elif not isinstance(children, list):
            if isinstance(children, str):
                children = [ children ]
            else:
                children = list(children)

        children = [ c if isinstance(c, RegexPattern) else NonCapturing(c) for c in children ]
        for c in children:
            c.add_parent(self)
        self._children = children
        self._child_map = None
        if parents is None:
            parents = []
        elif not isinstance(parents, list):
            parents = list(parents)
        self._parents = parents # we need this so we can propagate recompile calls all the way up
        self._joiner = joiner
        if join_function is None:
            def join_function(j, k, **ignore):
                try:
                    return j.join(k)
                except TypeError:
                    pass
                raise ValueError("failed to join {} with {}".format(k, j))
        elif join_function == "safe":
            join_function = self._join_kids
        self._join_function = join_function # sometimes we might want to join the kids in a slightly different way...
        self._wrapper_function = wrapper_function
        self._combine_args = ()
        self._combine_kwargs = {}

        self._prefix = prefix
        self._suffix = suffix
        self.parser = parser
        self.handler = handler
        self.default_value = default_value

        self._capturing = capturing
        self.allow_inner_captures = allow_inner_captures
        self.has_named_child = any((c.has_named_child or c.key is not None) for c in self._children)
        # since I chose to let named groups take precedence over regular groups I guess I need to be cognizant of this...
        self.has_capturing_child = any((c.has_capturing_child or c.capturing) for c in self._children)
        self._capturing_groups = None



    @property
    def pat(self):
        return self._pat
    @pat.setter
    def pat(self, pat):
        self._pat = pat
        self.invalidate_cache()
    @property
    def children(self):
        """

        :return:
        :rtype: tuple[RegexPattern]
        """
        return tuple(self._children)
    @property
    def child_count(self):
        """

        :return:
        :rtype: int
        """
        return len(self._children)
    @property
    def child_map(self):
        """Returns the map to subregexes for named regex components

        :return:
        :rtype: Dict[str, RegexPattern]
        """
        # we'll make a new OD every time to decrease the risk of mutability breakages
        # if self._child_map is None:
        #     self._child_map = OrderedDict( (r.key, r) for r in self._children if r.key is not None )
        return OrderedDict( (r.key, r) for r in self._children if r.key is not None )
    @property
    def parents(self):
        """

        :return:
        :rtype: tuple[RegexPattern]
        """
        return tuple(self._parents)
    @property
    def joiner(self):
        """

        :return:
        :rtype: str
        """
        return self._joiner
    @joiner.setter
    def joiner(self, j):
        self._joiner = j
        self._cached = None

    @property
    def join_function(self):
        """

        :return:
        :rtype: function
        """
        return self._join_function

    @join_function.setter
    def join_function(self, j):
        self._join_function = j
        self._cached = None
    @property
    def suffix(self):
        """

        :return:
        :rtype: str | RegexPattern
        """
        return self._suffix
    @suffix.setter
    def suffix(self, e):
        self._suffix = e
        self._cached = None
    @property
    def prefix(self):
        """

        :return:
        :rtype: str | RegexPattern
        """
        return self._prefix
    @prefix.setter
    def prefix(self, s):
        self._prefix = s
        self._cached = None
    @property
    def dtype(self):
        """Returns the StructuredType for the matched object

        The basic thing we do is build the type from the contained child dtypes
        The process effectively works like this:
            If there's a single object, we use its dtype no matter what
            Otherwise, we add together our type objects one by one, allowing the StructuredType to handle the calculus

        After we've built our raw types, we compute the shape on top of these, using the assigned repetitions object
        One thing I realize now I failed to do is to include the effects of sub-repetitions... only a single one will
        ever get called.

        :return:
        :rtype: None | StructuredType
        """
        dt = self._dtype

        # might be worth introducing some level of caching for this dtype so that we're not recomputing it
        # over and over for recursive processes
        if dt is None and self.child_count > 0:
            subdts = self.get_capturing_groups(allow_inners=False) # the recursive dtype computation will pick this up
            if len(subdts) == 0:
                subdts = self._children
            named = OrderedDict((g.key, g) for g in subdts if g.key is not None)
            if len(named) > 0:
                subdts = OrderedDict((k, g.dtype) for k, g in named.items())
            else:
                subdts = [ g.dtype for g in subdts ]

            if len(subdts) == 0:
                dt = StructuredType(str, default_value=self.default_value)
            elif isinstance(subdts, OrderedDict):
                dt = StructuredType(subdts, default_value=self.default_value)
            elif len(subdts) == 1:
                dt = subdts[0] # singular type
                if not isinstance(dt, StructuredType):
                    dt = StructuredType(dt, default_value=self.default_value)
            else:
                import functools as ft
                from operator import add

                if not isinstance(subdts[0], StructuredType):
                    subdts[0] = StructuredType(subdts[0], default_value=self.default_value)
                dt = ft.reduce(add, subdts) # build the compound type
            if self.repetitions is not None:
                reps = self.repetitions
                if isinstance(reps, int):
                    reps = (reps, None)
                dt = dt.repeat(*reps)
        elif dt is not None and (not isinstance(dt, StructuredType)):
            dt = StructuredType(dt, default_value=self.default_value)

        return dt

    @property
    def is_repeating(self):
        return isinstance(self.repetitions, tuple)
    @property
    def capturing(self):
        return self._capturing or (self._capturing is None and self.is_repeating and self.has_capturing_child)
    @capturing.setter
    def capturing(self, cap):
        self._capturing = cap

    def get_capturing_groups(self, allow_inners = None):
        """
        We walk down the tree to find the children with capturing groups in them and
        then find the outermost RegexPattern for those unless allow_inners is on in which case we pull them all
        """
        groups = []
        if allow_inners is None:
            allow_inners = self.allow_inner_captures
        for g in self._children:
            if g.capturing:
                groups.append(g)
                if allow_inners and g.has_capturing_child:
                    groups.extend(g.get_capturing_groups(allow_inners = True))
            elif g.has_capturing_child:
                groups.extend(g.get_capturing_groups(allow_inners = allow_inners))
        return groups

    @property
    def captures(self):
        """Subtly different from capturing n that it will tell us if we need to use the group in post-processing, essentially

        :return:
        :rtype:
        """

        return self.capturing or self.has_capturing_child or self.has_named_child
    @property
    def capturing_groups(self):
        """Returns the capturing children for the pattern

        :return:
        :rtype:
        """

        if not self.capturing and not self.has_capturing_child:
            return None
        elif not self.capturing:
            # we walk down the tree at this point, finding the outer-most capturing groups in a flat structure
            # I'd make it a tree but regex just returns captured stuff, it doesn't make a tree out of them
            if self._capturing_groups is None:
                self._capturing_groups = tuple(self.get_capturing_groups(allow_inners=self.allow_inner_captures))
            return self._capturing_groups
        else:
            return self

    @property
    def named_groups(self):
        """Returns the named children for the pattern

        :return:
        :rtype:
        """

        named = (self.key is not None)
        if not named and not self.has_named_child:
            return None
        elif not named:
            # we walk down the tree at this point, finding the outer-most named capturing groups in a flat structure
            # I'd make it a tree but regex just returns captured stuff, it doesn't make a tree out of them
            caps = self.capturing_groups
            if caps is None or caps is self:
                groups = None
            else:
                groups = OrderedDict(
                    (g.key, g) for g in caps if g.key is not None
                )
                if len(groups) == 0:
                    groups = None
            return groups
        else:
            return OrderedDict(((self.key, self), ) )

    # we want a Regex component to be reusable in lots of different Regexes without having to copy all of them
    # so we'll make it so that we can only walk _down_ a regex tree, and not worry about having to go back up
    # @property
    # def siblings(self):
    #     return

    def combine(self, other, *args, **kwargs):
        """Combines self and other

            :param other:
            :type other: RegexPattern | str
            :return:
            :rtype: str | callable
            """
        op = other.pat if isinstance(other, RegexPattern) else other
        if isinstance(self.pat, str) and isinstance(op, str):
            return self.pat + op
        elif isinstance(self.pat, str):
            def apply_op(p, *arg, wrap = op, prefix = self.pat, a = args, kw = kwargs, **kwarg):
                return prefix + wrap(p, *a, **kw)
            return apply_op
        elif isinstance(op, str):
            return self.pat(op, *args, **kwargs)
        else:
            def merge_ops(p, *arg, wrap = op, prewrap = self.pat, a = args, kw = kwargs, **kwarg):
                return prewrap(wrap(p, *a, **kw))
            return merge_ops

    def wrap(self, *args, **kwargs):
        """
        Applies wrapper function
        """
        self._combine_args = args
        self._combine_kwargs = kwargs
        if self._wrapper_function is not None:
            self._wrapper_function(self, *args, **kwargs)

    @staticmethod
    def _join_kids(joiner, kids, no_capture=True):
        if len(kids) > 1:
            kids = [ group(k, no_capture=no_capture) if not is_grouped(k) else k for k in kids ]
        return joiner.join(kids)
    def build(self,
              joiner = None,
              prefix = None,
              suffix = None,
              recompile = True,
              no_captures = False,
              verbose = False
              ):
        # might want to add a flag that checks if a block has already been wrapped in no-capture? That would cut down
        # on regex length...
        if recompile or self._cached is None:
            # this recompile rule might be excessive, but to be honest these compilations are so cheap it wouldn't even really
            # matter if recompile were always True
            recomp_captures = self.capturing and not self.allow_inner_captures
            no_caps = no_captures or recomp_captures
            kids = [
                c.build(
                    recompile = True if recompile else ('No Cache' if (c.has_capturing_child and recomp_captures) else False),
                    no_captures = no_caps
                ) for c in self._children
            ]

            if verbose:
                print("Compiling regex from: {}".format(kids))

            if joiner is None:
                joiner = self.joiner
            if prefix is None:
                prefix = self.prefix
            if suffix is None:
                suffix = self.suffix
            joiner = joiner.build( no_captures = no_caps ) if isinstance(joiner, RegexPattern) else joiner if joiner is not None else ""
            prefix = prefix.build( no_captures = no_caps ) if isinstance(prefix, RegexPattern) else prefix if prefix is not None else ""
            suffix = suffix.build( no_captures = no_caps ) if isinstance(suffix, RegexPattern) else suffix if suffix is not None else ""
            if verbose:
                print("Joiner:{}\n Prefix:{}\n Suffix:{}".format(joiner, prefix, suffix))

            # the big question now is how do I figure out if a Capturing was applied...?
            # I guess since I'm walking _down_ the tree I don't even need to check for that
            # I can just check self.capturing... huh wow
            if no_captures:
                # I guess we temporarily make our pattern a non-capturing one...?

                kwargs = self._combine_kwargs.copy()
                kwargs['no_capture'] = True

                comp = self.combine(
                    #unclear whether I should be putting the prefix/suffix inside or outside this ._.
                    prefix + self.join_function(joiner, kids, no_capture=True) + suffix,
                    *self._combine_args,
                    **kwargs
                )
            else:
                if 'no_capture' in self._combine_kwargs:
                    kwargs = self._combine_kwargs.copy()
                    kwargs['no_capture'] = (not self.capturing)
                else:
                    kwargs = self._combine_kwargs
                comp = self.combine(
                    #unclear whether I should be putting the prefix/suffix inside or outside this ._.
                    prefix + self.join_function(joiner, kids, no_capture=(not self.capturing)) + suffix,
                    *self._combine_args,
                    **kwargs
                )
            if isinstance(comp, RegexPattern): # to be honest I don't know how we got here...?
                comp = comp.build(
                    recompile= True if recompile else ('No Cache' if (comp.has_capturing_child and recomp_captures) else False),
                    no_captures = no_caps # is this right...?
                )

            if verbose:
                print("End Regex:", comp)

            if no_captures: # basically no_captures means operate in a not-quite-right world?
                return comp

            self._cached = comp

        return self._cached
    @property
    def compiled(self):
        if self._comp is None:
            self._comp = re.compile(self.build())
        return self._comp

    def add_parent(self, parent):
        self._parents.append(parent)
    def remove_parent(self, parent):
        self._parents.remove(parent)

    def add_child(self, child):
        self._children.append(child)
        self._child_map = None
        self.has_capturing_child = self.has_capturing_child or (child.capturing or child.has_capturing_child)
        self.has_named_child = self.has_named_child or (child.has_named_child or child.key is not None)
        self.invalidate_cache()
    def add_children(self, children):
        self._children.extend(children)
        self._child_map = None
        self.has_capturing_child = self.has_capturing_child or any((c.capturing or c.has_capturing_child) for c in children)
        self.has_named_child = self.has_named_child or any((c.has_named_child or c.key is not None) for c in children)
        self.invalidate_cache()
    def remove_child(self, child):
        self._children.remove(child)
        self._child_map = None
        self.has_capturing_child = self.has_capturing_child and any((c.capturing or c.has_capturing_child) for c in self._children)
        self.has_named_child = self.has_named_child and any((c.has_named_child or c.key is not None) for c in self._children)
        self.invalidate_cache()
    def insert_child(self, index, child):
        self._children.insert(index, child)
        self._child_map = None
        self.invalidate_cache()

    def invalidate_cache(self):
        # propagating this up the tree is the _only_ time I need _parents
        # which is a potential source of memory leaks... should I kill off the upwards propagation?
        self._cached = None
        self._comp = None
        self._capturing_groups = None
        for p in self._parents:
            p.invalidate_cache()

    # turns out python is clever enough to automatically handle cases like this?
    # def __del__(self):
    #     # we need this for cleanup to avoid memory leaks when people call
    #     # del ... on the Regex -- it won't work perfectly, but it'll be good enough
    #     for c in self._children:
    #         c.remove_parent(self)
    def __copy__(self):
        from copy import copy
        cls = type(self)
        new = cls.__new__(cls) # avoid init call
        new.__dict__.update(self.__dict__)
        new._children = copy(new._children) # copy the children list
        new._parents = [] # since this is a copy it has no parents
        new._cached = None # just safer...
        return new

    def __add__(self, other):
        """Combines self and other

        :param other:
        :type other: RegexPattern
        :return:
        :rtype:
        """
        if not isinstance(other, RegexPattern):
            other = RegexPattern(other)
        def null_pat(p):
            return p
        return type(self)(
            null_pat,
            children = [self, other]
        )
    def __radd__(self, other):
        """Combines self and other

        :param other:
        :type other: RegexPattern
        :return:
        :rtype:
        """
        if not isinstance(other, RegexPattern):
            other = RegexPattern(other)
        def null_pat(p):
            return p
        return type(self)(
            null_pat,
            children = [other, self]
        )
    def __call__(self, other,
                 *args,
                 # we'll treat this as an alternate constructor block
                 name = None,
                 dtype = None,
                 repetitions = None,
                 key = None,
                 joiner = None,
                 join_function = None,
                 wrap_function = None,
                 suffix = None,
                 prefix = None,
                 multiline = None,
                 parser = None,
                 handler = None,
                 capturing = None,
                 default=None,
                 allow_inner_captures = None,
                 **kwargs
                 ):
        """Wraps self around other

        :param other:
        :type other: RegexPattern
        :return:
        :rtype:
        """
        from copy import copy
        new = copy(self)

        if name is not None:
            new.name = name
        if key is not None:
            new.key = key
        if joiner is not None:
            new._joiner = joiner
        if join_function is not None:
            new._join_function = join_function
        if dtype is not None:
            new._dtype = dtype
        if repetitions is not None:
            new.repetitions = repetitions
        if wrap_function is not None:
            new._wrapper_function = wrap_function
        if suffix is not None:
            new._suffix = suffix
        if prefix is not None:
            new._prefix = prefix
        if parser is not None:
            new.parser = parser
        if handler is not None:
            new.handler = handler
        if default is not None:
            new.default_value = default
        if capturing is not None:
            new.capturing = capturing
        if allow_inner_captures is not None:
            new.allow_inner_captures = allow_inner_captures

        if isinstance(other, (tuple, list)):
            other = [ RegexPattern(o) if not isinstance(o, RegexPattern) else o for o in other ]
            new.add_children(other)
        else:
            if not isinstance(other, RegexPattern):
                other = RegexPattern(other)
            new.add_child(other)
        
        # print(self._wrapper_function, args, kwargs)
        new.wrap(*args, **kwargs)

        return new

    def __repr__(self):
        return "{}(key: {!r}, children : <{}>, pattern : {})".format(
            type(self).__name__,
            self.key,
            len(self._children),
            self._pat
        )
    def __str__(self):
        wat = self.build()
        return wat

    def __getitem__(self, item):
        return self.child_map[item]

    ### Supporting the re mechanisms directly for convenience
    def match(self, txt):
        return re.match(self.compiled, txt)
    def search(self, txt):
        return re.search(self.compiled, txt)
    def findall(self, txt):
        return re.findall(self.compiled, txt)
    def finditer(self, txt):
        return re.finditer(self.compiled, txt)

######################################################################################################################
#
#                                       Concrete RegexPatterns to combine and use
#
#
def is_grouped(p):
    """Takes a string pattern and tries to check if it's already in a singular construct (usually grouped...)

    :param p: pattern
    :type p: str
    :return:
    :rtype:
    """

    # first we'll check if it's syntactically _probably_ grouped
    all_is_well = starts_right = ends_right = p.startswith('(')
    if starts_right:
        ends_right = p.endswith(")")
        all_is_well = starts_right and ends_right
        if all_is_well:
            # we now want to check that we have balanced numbers of ( and ) ...
            count_l = p.count('(')
            count_r = p.count(')')
            all_is_well = count_l == count_r
        if all_is_well:
            # and I guess further check that we after splitting by the first ) we have an _unbalanced_ form
            right_hand = p.split('(', 1)[1]
            count_l = right_hand.count('(')
            count_r = right_hand.count(')')
            all_is_well = count_l == count_r

    if not all_is_well:
        # we'll then check if we have a simple character escape
        starts_right = all_is_well = p.startswith("\\")
        if starts_right:
            all_is_well = len(p) == 2 # indicates we

    if not all_is_well:
        # we'll try to check if we have a simple character class
        starts_right = p.startswith('[')
        if starts_right:
            ends_right = p.endswith("]")
            if starts_right and ends_right:
                all_is_well = p.count('[') == 1 and p.count(']') == 1

    return all_is_well
def group(p, no_capture = False):
    if no_capture:
        return non_capturing(p)
    else:
        return r"("+p+r")"
def non_capturing(p, *a, **kw):
    return r"(?:"+p+r")"
def optional(p, no_capture = False):
    if not is_grouped(p):
        return non_capturing(p) + "?"
    else:
        return p + "?"
def alternatives(p, no_capture = False):
    if is_grouped(p):
        return p
    else:
        return group(p, no_capture=no_capture)
def shortest(p, no_capture = False):
    if not p.endswith("*") | p.endswith("+"):
        if not is_grouped(p):
            return non_capturing(p) + "*?"
        else:
            return p + "*?"
    else:
        return p + "?"
def repeating(p, min = 1, max = None, no_capture = False):
    if not is_grouped(p):
        p = non_capturing(p)
    if max is None:
        if min is None:
            base_pattern = p+"*"
        elif min == 1:
            base_pattern = p+"+"
        else:
            base_pattern = p +"{" + str(min) + ",}"
    elif min == max:
        base_pattern = p +"{" + str(min) + "}"
    else:
        if min is None:
            min = 0
        base_pattern = p +"{" + str(min) + "," + str(max) + "}"
    if no_capture:
        return base_pattern
    else:
        return grp_p(base_pattern)
def duplicated(p, num, riffle="", no_capture = False):
    if isinstance(riffle, RegexPattern):
        riffle = str(riffle)
    return riffle.join([p]*num)
def named(p, n, no_capture=False):
    if no_capture:
        return non_cap_p(p)
    else:
        return "(?P<"+n+">"+p+")"

# wrapper patterns
grp_p = group # capturing group
Capturing = RegexPattern(grp_p, "Capturing", capturing=True)
Capturing.__name__ ="Capturing"
Capturing.__doc__ = """
    Represents a capturing group in a RegexPattern
    """

non_cap_p = non_capturing # non-capturing group
NonCapturing = RegexPattern(non_cap_p, "NonCapturing", dtype=DisappearingType)
NonCapturing.__name__ ="NonCapturing"
NonCapturing.__doc__ = """
    Represents something that should not be captured in a RegexPattern
    """

op_p = optional # optional group
def opnb_p(p, no_capture=False):
    return r"(?:"+p+r")?" # optional non-binding group
Optional = RegexPattern(optional,
                        "Optional"
                        )
Optional.__name__ ="Optional"
Optional.__doc__ = """
    Represents something that should be optional in a RegexPattern
    """

Alternatives = RegexPattern(alternatives, joiner="|")
Alternatives.__name__ ="Alternatives"
Alternatives.__doc__ = """
    Represents a set of alternatives in a RegexPattern
    """

lm_p = repeating
Longest = RegexPattern(lm_p, "Longest")
Longest.__name__ = "Longest"
Longest.__doc__ = """
    Represents that the longest match of the enclosed pattern should be searched for
    """

sm_p = shortest
Shortest = RegexPattern(sm_p, "Shortest")
Shortest.__name__ = "Shortest"
Shortest.__doc__ = """
    Represents that the shortest match of the enclosed pattern should be searched for
    """

def wrap_repeats(self, min = None, max = None, no_capture=None):
    self.repetitions = (min, max)
Repeating = RegexPattern(repeating,
                         "Repeating",
                         wrapper_function=wrap_repeats
                         )
Repeating.__name__ = "Repeating"
Repeating.__doc__ = """
    Represents that the patten can be repeated
    """

def wrap_name(self, n):
    self.key = n
    if self.name is None:
        self.name = n
Named = RegexPattern(
    named,
    "Named",
    wrapper_function = wrap_name,
    capturing=True
)
Named.__name__ = "Named"
Named.__doc__ = """
    Represents a named group. These are _always_ captured, to the exclusion of all else.
    """

def wrap_duplicate_type(self, n, riffle = ""):
    dt = self._dtype
    if isinstance(dt, tuple):
        dt, shape = dt
    else:
        shape = None
    if shape is None:
        shape = (n,)
    else:
        shape = (n,) + shape
    self._dtype = (dt, shape)
Duplicated = RegexPattern(duplicated,
                          "Duplicated",
                          wrapper_function = wrap_duplicate_type
                          )
Duplicated.__name__ = "Duplicated"
Duplicated.__doc__ = """
    Represents an explicitly duplicated pattern
    """

pc_p = lambda p, no_capture = False: r"["+p+r"]" # pattern class
PatternClass = RegexPattern(pc_p, "PatternClass")
PatternClass.__name__ = "PatternClass"
PatternClass.__doc__ = """
    Represents a pattern class, for wrapping other patterns
"""

parened_p = lambda p, no_capture = False: r"\("+p+"\)"
Parenthesized = RegexPattern(parened_p, "Parenthesized")
Parenthesized.__name__ = "Parenthesized"
Parenthesized.__doc__ = """
    Represents that something should be wrapped in parentheses, not treated as Capturing
    """

# raw declarative patters
any_p = "."
Any = RegexPattern(any_p, "Any")
Any.__name__ = "Any"
Any.__doc__ = """
    Represents any character
    """

sign_p = r"[\+\-]"
Sign = RegexPattern(sign_p, "Sign")
Sign.__name__ = "Sign"
Sign.__doc__ = """
    Represents a +/- sign
    """

paren_p = r"\("+".*?"+"\)"

num_p = opnb_p(sign_p)+r"\d*\.\d+" # real number
Number = RegexPattern(num_p, "Number", dtype=float)
Number.__name__ = "Number"
Number.__doc__ = """
    Represents a real number, like -1.23434; doesn't support "E" notation
    """

int_p = opnb_p(sign_p)+r"\d+" # integer
Integer = RegexPattern(int_p, "Integer", dtype=int)
Integer.__name__ = "Integer"
Integer.__doc__ = """
    Represents an integer
    """

posint_p = r"\d+" # only positive integer
PositiveInteger = RegexPattern(posint_p, "PositiveInteger", dtype=int)
PositiveInteger.__name__ = "PositiveInteger"
PositiveInteger.__doc__ = """
    Represents a positive integer (i.e. just a string of digits)
    """

ascii_p = "[a-zA-Z]"
ASCIILetter = RegexPattern(ascii_p, "ASCIILetter", dtype=str)
ASCIILetter.__name__ = "ASCIILetter"
ASCIILetter.__doc__ = """
    Represents a single ASCII letter
    """

name_p = ascii_p+"{1,2}" # atom name
AtomName = RegexPattern(name_p, "AtomName", dtype=str)
AtomName.__name__ = "AtomName"
AtomName.__doc__ = """
    Represents an atom symbol like Cl or O (this is misnamed, I know)
    """

ws_char_class = r"(?!\n)\s" # probably excessive... but w/e I'm not winning awards for speed here
WhitespaceCharacter = RegexPattern(ws_char_class, "WhitespaceCharacter", dtype=str)
WhitespaceCharacter.__name__ = "WhitespaceCharacter"
WhitespaceCharacter.__doc__ = """
    Represents a single whitespace character
    """

ws_p = non_capturing(ws_char_class)+"*" # whitespace
wsr_p = non_capturing(ws_char_class)+"+" # real whitespace
Whitespace = RegexPattern(ws_p, "Whitespace", dtype=str)
Whitespace.__name__ = "WhitespaceCharacter"
Whitespace.__doc__ = """
    Represents a block of whitespace
    """

WordCharacter = RegexPattern("\w", "WordCharacter", dtype=str)
WordCharacter.__name__ = "WordCharacter"
WordCharacter.__doc__ = """
    Represents a single number or letter (i.e. non-whitespace)
    """

Word = RegexPattern("\w+", "Word", dtype=str)
Word.__name__ = "Word"
Word.__doc__ = """
    Represents a block of WordCharacters
    """

VariableName = RegexPattern((ASCIILetter, Word), joiner="", dtype=str)
VariableName.__name__ = "VariableName"
VariableName.__doc__ = """
    Represents a possible variable name sans underscored, basically an ASCIILetter and then a word
    """

ascii_punc_char_class = r"\.,<>?/'\";:{}\[\]\+=\(\)\*&\^%$#@!~`"
ASCIIPunctuation = RegexPattern(ascii_punc_char_class, "ASCIIPunctuation", dtype=str)
ASCIIPunctuation.__name__ = "ASCIIPunctuation"
ASCIIPunctuation.__doc__ = """
    Represents a single piece of punctuation
    """

cart_p = ws_p.join([ grp_p(num_p) ]*3) # cartesian coordinate
CartesianPoint = RegexPattern(cart_p, "CartesianPoint", dtype=(float, (3,)))
CartesianPoint.__name__ = "CartesianPoint"
CartesianPoint.__doc__ = """
    Represents a 'point', i.e. 3 numbers separated by whitespace
    """

acart_p = "("+int_p+")"+ws_p+cart_p # atom coordinate as comes from a XYZ table
IntXYZLine = RegexPattern(acart_p, "IntXYZLine", dtype=(int, (float, (3,))))
IntXYZLine.__name__ = "IntXYZLine"
IntXYZLine.__doc__ = """
    Represents a line in an XYZ file that starts with an int, like
    ```
    1   -1.232323 2.23212421 43.44343434
    ```
    """

aNcart_p = "("+name_p+")"+ws_p+cart_p # atom coordinate as comes from a XYZ table
XYZLine = RegexPattern(aNcart_p, "XYZLine", dtype=(str, (float, (3,))))
XYZLine.__name__ = "XYZLine"
XYZLine.__doc__ = """
    Represents a line in an XYZ file that starts with an atom name, like
    ```
    Cl   -1.232323 2.23212421 43.44343434
    ```
    """

Empty = RegexPattern("", "Empty")
Empty.__name__ = "Empty"
Empty.__doc__ = """
    Represents an empty pattern...I can't remember why this is here
    """

Newline = RegexPattern(r"\n", "Newline", dtype=str)
Newline.__name__ = "Newline"
Newline.__doc__ = """
    Represents a newline character
    """

ZMatPattern = Capturing(AtomName)
for i in range(3):
    ZMatPattern += Optional(NonCapturing(
        Repeating(Whitespace, 1) + Capturing(PositiveInteger) + # ref int
        Repeating(Whitespace, 1) + Capturing(Number) # ref value
    ))
ZMatPattern.name = "ZMatPattern"
ZMatPattern.__name__ = "ZMatPattern"
ZMatPattern.__doc__ = """
    Represents Z-matrix block
    """





__all__ = [
    "RDMolecule"
]

import numpy as np, io, os
from .. import Numputils as nput

from .ChemToolkits import RDKitInterface

class RDMolecule:
    """
    A simple interchange format for RDKit molecules
    """

    def __init__(self, rdconf, charge=None):
        #atoms, coords, bonds):
        self.conf = rdconf
        self.charge = charge

    @property
    def rdmol(self):
        return self.conf.GetOwningMol()
    @property
    def atoms(self):
        mol = self.rdmol
        return [atom.GetSymbol() for atom in mol.GetAtoms()]
    @property
    def bonds(self):
        mol = self.rdmol
        return [
            [b.GetBeginAtomIdx(), b.GetEndAtomIdx(), b.GetBondTypeAsDouble()]
            for b in mol.GetBonds()
        ]
    @property
    def coords(self):
        return self.conf.GetPositions()
    @property
    def rings(self):
        return self.rdmol.GetRingInfo().AtomRings()
    @property
    def meta(self):
        return self.rdmol.GetPropsAsDict()

    @classmethod
    def chem_api(cls):
        return RDKitInterface.submodule("Chem")
    @classmethod
    def from_rdmol(cls, rdmol, conf_id=0, charge=None, guess_bonds=False, sanitize=True, sanitize_ops=None):
        if guess_bonds:
            Chem = cls.chem_api() # to get nice errors
            rdDetermineBonds = RDKitInterface.submodule("Chem.rdDetermineBonds")
            rdmol = Chem.Mol(rdmol)
            if charge is None:
                charge = 0
            rdDetermineBonds.DetermineConnectivity(rdmol, charge=charge)
            # return cls.from_rdmol(rdmol, conf_id=conf_id, guess_bonds=False, charge=charge)
        if sanitize:
            Chem = cls.chem_api() # to get nice errors
            rdmolops = RDKitInterface.submodule("Chem.rdmolops")
            if sanitize_ops is None:
                sanitize_ops = (
                        rdmolops.SANITIZE_ALL
                        ^rdmolops.SANITIZE_PROPERTIES
                        # ^rdmolops.SANITIZE_ADJUSTHS
                        # ^rdmolops.SANITIZE_CLEANUP
                        ^rdmolops.SANITIZE_CLEANUP_ORGANOMETALLICS
                )
            rdmol = Chem.Mol(rdmol)
            Chem.SanitizeMol(rdmol, sanitize_ops)
        conf = rdmol.GetConformer(conf_id)
        return cls(conf, charge=charge)

    @classmethod
    def from_coords(cls, atoms, coords, bonds=None, charge=None, guess_bonds=None):
        Chem = cls.chem_api()
        mol = Chem.EditableMol(Chem.Mol())
        mol.BeginBatchEdit()
        for a in atoms:
            a = Chem.Atom(a)
            mol.AddAtom(a)
        if bonds is not None:
            for b in bonds:
                if len(b) == 2:
                    i,j = b
                    t = 1
                else:
                    i,j,t = b
                if nput.is_numeric(t):
                    t = Chem.BondType.values[int(t)]
                else:
                    t = Chem.BondType.names[t]
                mol.AddBond(i, j, t)
        mol.CommitBatchEdit()

        mol = mol.GetMol()
        conf = Chem.Conformer(len(atoms))
        conf.SetPositions(np.asanyarray(coords))
        conf.SetId(0)
        mol.AddConformer(conf)

        if guess_bonds is None:
            guess_bonds = bonds is None

        return cls.from_rdmol(mol, charge=charge, guess_bonds=guess_bonds)

    @classmethod
    def from_mol(cls, mol, coord_unit="Angstroms", guess_bonds=None):
        from ..Data import UnitsData

        return cls.from_coords(
            mol.atoms,
            mol.coords * UnitsData.convert(coord_unit, "Angstroms"),
            bonds=mol.bonds,
            charge=mol.charge,
            guess_bonds=guess_bonds
        )

    @classmethod
    def _load_sdf_conf(cls, stream, which=0):
        Chem = cls.chem_api()
        mol = None
        for i in range(which+1):
            mol = next(Chem.ForwardSDMolSupplier(stream, sanitize=False, removeHs=False))
        return mol
    @classmethod
    def from_sdf(cls, sdf_string, which=0):
        if os.path.isfile(sdf_string):
            with open(sdf_string, 'rb') as stream:
                mol = cls._load_sdf_conf(stream, which=which)
        else:
            mol = cls._load_sdf_conf(io.BytesIO(sdf_string.encode()), which=which)
        return cls.from_rdmol(mol)

    @classmethod
    def get_confgen_opts(cls):
        return {}
    @classmethod
    def from_smiles(cls, smiles, num_confs=1, add_implicit_hydrogens=False):

        if os.path.isfile(smiles):
            with open(smiles) as f:
                smiles = f.read()
        Chem = cls.chem_api()
        params = Chem.SmilesParserParams()
        params.removeHs = False
        rdkit_mol = Chem.MolFromSmiles(smiles, params)
        mol = Chem.AddHs(rdkit_mol, explicitOnly=not add_implicit_hydrogens)
        rdDistGeom = RDKitInterface.submodule("Chem.rdDistGeom")
        rdDistGeom.EmbedMolecule(mol, num_confs, **cls.get_confgen_opts())

        return cls.from_rdmol(mol)
    @classmethod
    def from_molblock(cls, molblock):
        Chem = cls.chem_api()
        if os.path.isfile(molblock):
            mol = Chem.MolFromMolFile(molblock, sanitize=False, removeHs=False)
        else:
            mol = Chem.MolFromMolBlock(molblock, sanitize=False, removeHs=False)
        return cls.from_rdmol(mol)





"""
Provides development utilities.
Each utility attempts to be almost entirely standalone (although there is
a small amount of cross-talk within the packages).
In order of usefulness, the design is:
1. `Logging` provides a flexible logging interface where the log data can be
    reparsed and loggers can be passed around
2. `Serializers`/`Checkpointing` provides interfaces for writing/loading data
    to file and allows for easy checkpoint loading
3. `Jobs` provides simpler interfaces for running jobs using the existing utilities
4. `CLIs` provides simple command line interface helpers
"""

__all__ = []
from .Caches import *; from .Caches import __all__ as exposed
__all__ += exposed
from .Serializers import *; from .Serializers import __all__ as exposed
__all__ += exposed
from .Schema import *; from .Schema import __all__ as exposed
__all__ += exposed
from .Logging import *; from .Logging import __all__ as exposed
__all__ += exposed
from .Checkpointing import *; from .Checkpointing import __all__ as exposed
__all__ += exposed
from .Persistence import *; from .Persistence import __all__ as exposed
__all__ += exposed
from .ObjectBackers import *; from .ObjectBackers import __all__ as exposed
__all__ += exposed
from .Configurations import *; from .Configurations import __all__ as exposed
__all__ += exposed
from .Jobs import *; from .Jobs import __all__ as exposed
__all__ += exposed
from .CLIs import *; from .CLIs import __all__ as exposed
__all__ += exposed


"""
Provides a QuantityArray class to manage data & units simultaneously
"""

import numpy as np, io

__all__ = [
    "QuantityArray"
]

class QuantityArrayException(Exception):
    ...

class QuantityArray:
    """
    A little helper for working with NumPy arrays with units.
It's mostly just a safety mechanism for imported data, but also helps keep you from messing up when you
  do addition and multiplication and stuff.
    """
    def __init__(self, array, units):
        """
        :param array: array data
        :type array: np.ndarray
        :param unit: list of units for the array
        :type unit: str | Iterable[str]
        """
        if isinstance(units, str):
            units = [units]
        if len(units) != self.array.ndim:
            raise QuantityArrayException("units {} can't correspond to an array of shape {}".format(
                units,
                array.shape
            ))
        self.array = array
        self.units = tuple(units)
    @property
    def shape(self):
        return self.array.shape
    @property
    def dtype(self):
        return self.array.dtype

    @classmethod
    def raise_unit_mismatch(cls, u1, u2):
        raise QuantityArrayException("units '{}' and '{}' are incompatible".format(
            u1, u2
        ))
    def __neg__(self):
        """Implements -a"""
        return type(self)(-self.array, self.units)
    def __pos__(self):
        """Implements +a"""
        return type(self)(+self.array, self.units)
    def __add__(self, other):
        """Implements a+b"""
        if isinstance(other, (np.array, int, float, np.integer, np.floating)):
            return type(self)(other + self.array, self.units)
        elif isinstance(other, QuantityArray):
            if other.units != self.units:
                self.raise_unit_mismatch(self.units, other.units)
            return self.__add__(other.array)
        else:
            raise TypeError("unsupported operand type(s) for +: '{}' and '{}'".format(
                type(self),
                type(other)
            ))
    def __sub__(self, other):
        """Implements a-b"""
        if isinstance(other, (np.array, int, float, np.integer, np.floating)):
            return type(self)(other - self.array, self.units)
        elif isinstance(other, QuantityArray):
            if other.units != self.units:
                self.raise_unit_mismatch(self.units, other.units)
            return self.__mul__(other.array)
        else:
            raise TypeError("unsupported operand type(s) for -: '{}' and '{}'".format(
                type(self),
                type(other)
            ))
    def __mul__(self, other):
        """Implements a*b"""
        if isinstance(other, (np.array, int, float, np.integer, np.floating)):
            return type(self)(other * self.array, self.units)
        elif isinstance(other, QuantityArray):
            if other.units != self.units:
                self.raise_unit_mismatch(self.units, other.units)
            return self.__mul__(other.array)
        else:
            raise TypeError("unsupported operand type(s) for *: '{}' and '{}'".format(
                type(self),
                type(other)
            ))
    def __truediv__(self, other):
        """Implements a/b"""
        if isinstance(other, (np.array, int, float, np.integer, np.floating)):
            return type(self)(other / self.array, self.units)
        elif isinstance(other, QuantityArray):
            if other.units != self.units:
                self.raise_unit_mismatch(self.units, other.units)
            return self.__mul__(other.array)
        else:
            raise TypeError("unsupported operand type(s) for /: '{}' and '{}'".format(
                type(self),
                type(other)
            ))
    def __divmod__(self, other):
        """Implements a//b"""
        if isinstance(other, (np.array, int, float, np.integer, np.floating)):
            return type(self)(other // self.array, self.units)
        elif isinstance(other, QuantityArray):
            if other.units != self.units:
                self.raise_unit_mismatch(self.units, other.units)
            return self.__mul__(other.array)
        else:
            raise TypeError("unsupported operand type(s) for //: '{}' and '{}'".format(
                type(self),
                type(other)
            ))
    def convert(self, units):
        """
        Converts the array from units A to units B
        :param units:
        :type units:
        :return:
        :rtype:
        """
        from ..Data import UnitsData

        if isinstance(units, str):
            units = [units]
        units = tuple(units)
        if units == self.units:
            return self
        if len(units) != len(self.units):
            raise QuantityArrayException("{}.{}: can't convert from units {} to {}".format(
                type(self).__name__,
                'convert',
                units,
                self.units
            ))

        factor = 1
        for old, new in zip(self.units, units):
            if old != new:
                factor *= UnitsData.convert(old, new)

        return type(self)(self.array*factor, units)
    def save(self, file):
        """
        Saves the QuantityArray to a file

        :param file:
        :type file:
        :return:
        :rtype:
        """
        np.savez(file, array=self.array, units=np.array(self.units, dtype=str))
    @classmethod
    def load(cls, file):
        """
        Loads a QuantityArray from file

        :param file:
        :type file:
        :return:
        :rtype:
        """
        dat = np.load(file)
        return cls(dat['array'], dat['units'])
    def format_header(self):
        return "Units: {} | Shape: {}".format(
            self.units,
            self.shape
        )
    @classmethod
    def parse_header(cls, line):
        unit, shape = line.split("|")
        # this kind of thing should never be done in legit production
        # code since it's fundamentally insecure...but we also do
        # dynamic loading of modules which is _also_ fundamentally insecure
        # so like we're already garbage
        unit = eval(unit.strip("Units: "))
        shape = eval(shape.strip("Shape: "))
        return (unit, shape)
    def savetxt(self, file):
        """
        Saves the QuantityArray to a text file
        :param file:
        :type file: file
        :return:
        :rtype:
        """
        opened_file = False
        try:
            if isinstance(file, str):
                file = open(file, "w")
                opened_file = True
            file.writelines([self.format_header()])
            array = self.array
            if self.array.ndim > 2:
                new_shape = (np.prod(array.shape[:-2]), array.shape[-1])
                array = array.reshape(new_shape)
            np.savetxt(file, array)
        finally:
            if opened_file:
                file.close()
    @classmethod
    def loadtxt(cls, file):
        """
        Loads a QuantityArray from a text file

        :param file:
        :type file:
        :return:
        :rtype:
        """
        opened_file = False
        try:
            if isinstance(file, str):
                file = open(file, "r")
                opened_file = True
            line_0 = file.readline()
            units, shape = cls.parse_header(line_0)
            array = np.loadtxt(file)
            return cls(array.reshape(shape), units)
        finally:
            if opened_file:
                file.close()
    def __repr__(self):
        return "{}(units={}, shape={}, dtype={})".format(
            type(self).__name__,
            self.units,
            self.shape,
            self.dtype
        )





"""
Simple package for easily creating command line interfaces in a
nestable way with automatic argument dispatch
"""

import inspect, typing, os, sys, argparse, collections

__all__ = [
    "CLI",
    "CommandGroup",
    "Command"
]

class Command:
    """
    A holder for a command that just automates type handling &
    that sort of thing
    """
    def __init__(self, name, method):
        self.name = name
        self.meth = method
        self.hints = typing.get_type_hints(method)
        self.pos_only_args = {
            arg.name for arg in inspect.signature(method).parameters.values()
            if arg.default is inspect.Parameter.empty
        }

    def get_help(self):
        """
        Gets single method help string
        :return:
        :rtype:
        """
        base_str = "{} - ".format(self.name)
        padding = " " * len(base_str)
        doc_str = self.meth.__doc__
        if doc_str is None:
            doc_str = ""
        doc_str = inspect.cleandoc(doc_str)
        doc_str += "\n" + "ARGS {}".format(
            " ".join(
                "{}{}:{}".format(
                    "*" if k in self.pos_only_args else "",
                    k, self._hint_str(v)
                ) for k, v in self.hints.items()
            )
        )
        doc_str = doc_str.strip()

        return base_str + doc_str.replace("\n", "\n" + padding)

    @classmethod
    def _hint_types(cls, hint):
        """
        :param hint:
        :type hint: type | str | typing.GenericAlias
        :return:
        :rtype:
        """
        if isinstance(hint, type):
            return (hint,)
        elif hasattr(hint, '__origin__') and hint.__origin__ is typing.Union:
            return tuple(a for a in hint.__args__ if a is not type(None))
        else:
            return None
    @classmethod
    def _hint_str(cls, hint):
        """
        :param hint:
        :type hint: type | str | typing.GenericAlias
        :return:
        :rtype:
        """
        if hint is type(None):
            hint = "None"
        elif isinstance(hint, type):
            hint = hint.__name__
        elif isinstance(hint, tuple):
            hint = " | ".join(cls._hint_str(x) for x in hint)
        elif isinstance(hint, str):
            hint = hint
        else:
            types = cls._hint_types(hint)
            if types is None:
                hint = str(hint).replace("typing.", "")
            else:
                return cls._hint_str(types)

        return hint

    @staticmethod
    def get_parse_dict(*spec):
        """
        Builds a parse spec to feed into an ArgumentParser later
        :param spec:
        :type spec:
        :return:
        :rtype:
        """
        argv_0 = sys.argv[0]
        try:
            sys.argv[0] = "parsing_dict"  # self.group + " " + self.cmd
            parser = argparse.ArgumentParser(add_help=False)
            keys = []
            for arg in spec:
                if len(arg) > 1:
                    arg_name, arg_dict = arg
                else:
                    arg_name = arg[0]
                    arg_dict = {}
                if 'dest' in arg_dict:
                    keys.append(arg_dict['dest'])
                else:
                    keys.append(arg_name)
                parser.add_argument(arg_name, **arg_dict)
            args = parser.parse_args()
            opts = {k: getattr(args, k) for k in keys}
        finally:
            sys.argv[0] = argv_0
        return {k: o for k, o in opts.items() if not (isinstance(o, str) and o == "")}

    @staticmethod
    def _get_typed(val:str, converter:callable):
        if val == "":
            return None
        else:
            return converter(val)
    def get_parse_spec(self):
        """
        Gets a parse spec that can be fed to ArgumentParser

        :return:
        :rtype:
        """
        args = []
        for k, v in self.hints.items():
            spec = {'default':""}
            if k not in self.pos_only_args:
                spec['dest'] = k
                key = "--" + k
            else:
                key = k
            types = self._hint_types(v)
            if types is None or len(types) == 0:
                spec['type'] = lambda v:v
            elif int in types:
                spec['type'] = lambda v: self._get_typed(v, int)
            elif float in types:
                spec['type'] = lambda v: self._get_typed(v, float)
            elif bool in types:
                spec['type'] = lambda v: self._get_typed(v, bool)
            elif str in types:
                spec['type'] = lambda v: v
            else:
                raise ValueError("don't know what to do with type hint {}".format(types))

            args.append((key, spec))
        return args

    def parse(self):
        """
        Generates a parse spec, builds an ArgumentParser, and parses the arguments

        :return:
        :rtype:
        """
        spec = self.get_parse_spec()
        return self.get_parse_dict(*spec)

    def __call__(self):
        """
        Parse argv and call bound method
        :return:
        :rtype:
        """
        parse = self.parse()
        args = []
        kwargs = {}
        for k in self.hints.keys():
            if k in self.pos_only_args:
                # if k not in parse:
                #     raise ValueError("positional only argument {} missing")
                args.append(parse[k])
            else:
                if k in parse:
                    kwargs[k] = parse[k]
        return self.meth(*args, **kwargs)

class CommandGroup:
    """
    Generic interface that defines an available set of commands
    as class methods.
    Basically just exists to be ingested by a CLI.
    """
    _name=""
    _description=""
    _tag=None # the group tag that the CLI will ingest

    @classmethod
    def _get_tag(cls):
        if cls._tag is None:
            return cls._name.lower()
        else:
            return cls._tag

    @classmethod
    def _get_commands(cls):
        """
        Returns the commands defined on the class as well
        as their arguments and allowed types
        :return:
        :rtype:
        """

        methods = {}
        for name, attr in cls.__dict__.items():
            if isinstance(name, str) and not name.startswith("_"):
                # print("...", name, inspect.ismethoddescriptor(attr))
                if inspect.ismethoddescriptor(attr):
                    methods[name] = Command(name, getattr(cls, name))
        sort_keys = sorted(methods.keys())
        sort_methods = collections.OrderedDict(
            (k, methods[k]) for k in sort_keys
        )

        return sort_methods

    @classmethod
    def _get_command(cls, k):
        """
        Returns the commands defined on the class as well
        as their arguments and allowed types
        :return:
        :rtype:
        """

        methods = cls._get_commands()
        if k in methods:
            return methods[k]
        else:
            raise KeyError("No command '{}' for group '{}'. Available commands: {}".format(k, cls._get_tag(), list(methods.keys())))

    @classmethod
    def _get_help(cls):
        """
        Gets the help string for the CLI
        :return:
        :rtype:
        """

        help_str = "\n  ".join(
            ["{}: {}".format(cls._get_tag(), cls._description)]
            + [c.get_help().replace("\n", "\n  ") for c in cls._get_commands().values()]
        )

        return help_str


class CLI:
    """
    A representation of a command line interface
    which layers simple command dispatching on the basic
    ArgParse interface
    """
    def __init__(self, name, description, *groups, cmd_name=None):
        """
        :param name:
        :type name: str
        :param name:
        :type description: str
        :param cmd_name:
        :type cmd_name: str | None
        :param groups:
        :type groups: Type[CommandGroup]
        """
        self.name = name
        self.description=description
        self.cmd_name = name.lower() if cmd_name is None else cmd_name
        self.groups = collections.OrderedDict(
            (g._get_tag(), g) for g in groups
        )
        self.argv = sys.argv

    def parse_group_command(self):
        """
        Parses a group and command argument (if possible) and prunes `sys.argv`

        :param group:
        :type group:
        :param command:
        :type command:
        :return:
        :rtype:
        """
        parser = argparse.ArgumentParser(add_help=False)
        parser.add_argument("group", type=str)
        parser.add_argument("command", type=str, default='', nargs="?")
        parse, unknown = parser.parse_known_args()
        group = parse.group
        cmd = parse.command
        sys.argv = [sys.argv[0]] + unknown

        return (group, cmd)

    def get_command(self):
        group, cmd = self.parse_group_command()
        if cmd == "": # either means we asked for --help or we've got a default command group
            if '--help' == sys.argv[0]: # quick hack but effective
                return self.get_group(cmd)._get_help()
            else:
                cmd, group = group, cmd
        return self.get_group(group)._get_command(cmd)

    def get_group(self, grp):
        if grp in self.groups:
            return self.groups[grp]
        elif grp == "":
            raise KeyError("No default command group (i.e. with key ''). Available groups: {}".format(list(self.groups.keys())))
        else:
            raise KeyError("No command group '{}'. Available groups: {}".format(grp, list(self.groups.keys())))

    def run_command(self):
        res = self.get_command()
        if not isinstance(res, str):
            # parse and call semantics are
            # handled by the stored Command object
            res = res()
        else:
            print(res)
        return res

    def get_help(self):
        """
        Gets the help string for the CLI
        :return:
        :rtype:
        """

        title_tag = "{} [{}]: ".format(self.cmd_name, "|".join(self.extra_commands.keys()))
        title_tag += self.description.replace("\n", "\n" + " " * len(title_tag))
        help_str = "\n".join([
            title_tag,
            *(g._get_help() for g in self.groups.values())
        ])

        return help_str

    def help(self, print_help=True):
        sys.argv.pop(1)
        res = self.get_help()
        if print_help:
            print(res)
        return res

    def run_parse(self, parse, unknown):
        """
        Provides a standard entry point to running stuff using the default CLI

        :param parse:
        :type parse:
        :param unknown:
        :type unknown:
        :return:
        :rtype:
        """

        # detect whether interactive run or not
        # interact = parse.interact or (
        #         len(sys.argv) == 1
        #         and not parse.help
        #         and not parse.script
        # )
        interact = parse.interact

        # in interactive/script envs we expose stuff
        if parse.script or interact:
            sys.path.insert(0, os.getcwd())
            interactive_env = {
                "__name__": "McEnv.script"
            }

        # In a script environment we just read in the script and run it
        # this is usefule because often these kinds of CLIs are set up
        # for things like containerized runtimes.
        # In that case it's nice to have a direct route into the loaded code
        # without having to set up some kind of backdoor
        if parse.script:
            script = sys.argv[1]
            sys.argv.pop(0)
            interactive_env["__name__"] = self.name + ".scripts." + os.path.splitext(os.path.basename(script))[0]
            with open(script) as scr:
                src = scr.read()
                src = compile(src, script, 'exec')
            interactive_env["__file__"] = script
            exec(src, interactive_env, interactive_env)
        elif parse.help:
            self.help(print_help=True)
        elif len(unknown) > 0:
            sys.argv = sys.argv + unknown
            self.run_command()

        # Start an interactive python...basically just because it can be convenient
        if interact:
            import code
            code.interact(banner=self.name + " Interactive Session", readfunc=None, local=interactive_env, exitmsg=None)

    extra_commands = {
        '--interact': dict(
            default=False, action='store_const', const=True, dest="interact",
            help='start an interactive session after running'
        ),
        '--script': dict(
            default=False, action='store_const', const=True, dest="script",
            help='run a script'
        ),
        '--help': dict(
            default=False, action='store_const', const=True, dest="help",
            help='print help string'
        ),
        '--fulltb': dict(
            default=False, action='store_const', const=True, dest="full_traceback",
            help='print full traceback'
        )
    }
    def parse_toplevel_args(self):
        """
        Parses out the top level flags that the program supports

        :return:
        :rtype:
        """

        parser = argparse.ArgumentParser(add_help=False)
        for k,v in self.extra_commands.items():
            parser.add_argument(k, **v)

        new_argv = []
        for k in sys.argv[1:]:
            if not k.startswith("--"):
                break
            new_argv.append(k)
        unknown = sys.argv[1 + len(new_argv):]
        sys.argv = [sys.argv[0]] + new_argv
        parse = parser.parse_args()

        return parse, unknown

    def run(self):
        """
        Parses the arguments in `sys.argv` and dispatches to the approriate action.
        By default supports interactive sessions, running scripts, and abbreviated tracebacks.

        :return:
        :rtype:
        """

        parse, unknown = self.parse_toplevel_args()
        if parse.full_traceback:
            self.run_parse(parse, unknown)
        else:
            error = None
            try:
                self.run_parse(parse, unknown)
            except Exception as e:
                error = e
            if error is not None:
                print(error)




"""
Provides a class that will walk through a set of objects & their children, as loaded into memory, and will generate Markdown for each.
The actual object Markdown is written by the things in the `Writers` module.
"""
import os, types, collections, inspect, importlib, re, uuid
import subprocess
import sys

from ..Formatters.TemplateEngine import *
from ..Misc import mixedmethod
from .ExamplesParser import ExamplesParser
from .MarkdownTemplates import MarkdownTemplateFormatter, MarkdownOps

__reload_hook__ = ["..Misc"]

__all__ = [
    "DocWalker",
    "ModuleWriter",
    "ClassWriter",
    "FunctionWriter",
    "MethodWriter",
    "ObjectWriter",
    "IndexWriter",
    "jdoc"
]

from ..Formatters.TemplateEngine.TemplateEngine import TemplateInterfaceList


class DocSpec(ObjectSpec):
    """
    A specification for an object to document.
    Supports the fields given by `spec_fields`.
    """
    spec_fields = (
        'id', # name for resolving the object
        'parent', # parent object name for writing,
        'children',
        'examples_root',
        'tests_root'
    )

    def __repr__(self):
        return '{}({})'.format(
            type(self).__name__,
            super().__repr__()
        )

class ExamplesExtractor(TemplateResourceExtractor):
    resource_keys = ['examples']
    resource_attrs = ['__examples__']

class TestsExtractor(TemplateResourceExtractor):
    resource_keys = ['tests']
    resource_attrs = ['__tests__']
    extension = 'Tests.py'
    def path_extension(self, handler:TemplateHandler):
        """
        Provides the default examples path for the object
        :return:
        :rtype:
        """
        base = super().path_extension(handler)
        return [base, os.path.split(base)[1]]
    def load(self, handler:TemplateHandler):
        res = super().load(handler)
        if res is not None:
            res = ExamplesParser(res)
        return res

class TestExamplesFormatter:
    def __init__(self, parser):
        if isinstance(parser, str):
            parser = ExamplesParser(parser)
        self.parser = parser

    @classmethod
    def from_file(cls, tests_file):
        with open(tests_file) as f:
            return cls(f.read())

    def get_template_parameters(self):
        """
        Formats an examples file

        :return:
        :rtype:
        """
        imports = self.parser.setup
        setup = "\n".join(self.parser.format_node(node) for node in imports)

        cls = self.parser.class_spec
        class_setup = "\n".join(
            [
                "class {}({}):".format(
                    cls[0].name,
                    ",".join(x.id for x in cls[0].bases)
                ),
                *(self.parser.format_node(node) for node in cls[1])
            ]
        )

        return dict(
            names=list(self.parser.functions.keys()),
            tests=[self.parser.format_node(v) for v in self.parser.functions.values()],
            setup=setup,
            class_setup=class_setup
        )

class DocTemplateOps(MarkdownOps):
    ...

class InteractiveTemplateEngine(TemplateInterfaceEngine):
    def __init__(self, templates:TemplateInterfaceList=None, ignore_missing=False, formatter_class=None, ignore_paths=()):
        if templates is None:
            templates = {
                'index.md':self.index_browser,
                'module.md':self.module_browser,
                'class.md':self.class_browser,
                'object.md':self.object_browser,
                'function.md':self.function_browser,
                'method.md':self.method_browser,
            }
        self.id = str(uuid.uuid4())
        self.pane = None
        super().__init__(templates,
                         ignore_missing=ignore_missing,
                         formatter_class=formatter_class,
                         ignore_paths=ignore_paths
                         )

    def clean_params(self, params):
        return {
            k: v for k, v in
            params.items()
            if not (v is None or isinstance(v, str) and len(v) == 0)
        }
    def prep_pars(self, writer, pars):
        from ..Jupyter import JHTML

        return {
            JHTML.SubsubsubHeading(k):
                (
                    JHTML.Markdown(v) if k in {'Details', 'Examples'}
                    else self.format_related_links(writer, v) if k in {'Related'}
                    else v
                )
                for k, v in pars.items()
        }
    def format_parameters_table(self, parameters):
        from ..Jupyter import JHTML, Flex

        return Flex(
            [
                JHTML.Div(
                    [
                        JHTML.Span(JHTML.Code(k), ": ", JHTML.Code(p['type'])),
                        " - ",
                        JHTML.Markdown(p["description"]).add_class('d-inline-block'),
                    ]
                )
                for k, p in parameters.items()
            ],
            direction='column'
        )
    def format_props_table(self, writer, props):
        from ..Jupyter import JHTML, Flex

        return Flex(
            [
                JHTML.Span(JHTML.Code(k), ": ", JHTML.Code(v.__name__ if isinstance(v, type) else type(v).__name__))
                for k, v in props
            ],
            direction='column'
        )

    def format_related_links(self, writer, related):
        from ..Jupyter import JHTML, LinkButton, Var

        _cache = set()
        def _get_docs(x):
            try:
                o = writer.resolve_relative_obj(x.strip())
                if o not in _cache:
                    _cache.add(o) # debounce
                return jdoc(o, engine=self)
            except Exception as e:
                import traceback
                return "\n".join(traceback.format_tb(sys.exc_info()[2])+[str(e)])

        return JHTML.List([
            JHTML.Li(
                LinkButton(
                    x.strip(),
                    lambda *_,x=x,**__:Var('current_pane', namespace=self.id).set_value(_get_docs(x))
                )
            ) for x in related.split(",")
        ])

    def index_browser(self,
                      index_files=None,
                      details=None,
                      related=None,
                      description=None,
                      examples=None,
                      _self=None,
                      **kw
                      ):
        from ..Jupyter import JHTML, Flex, Opener, VariableDisplay
        body = [
                JHTML.Markdown(description),
                *index_files
            ]
        params = self.clean_params({
                'Details': details,
                'Examples': examples,
                'Related': related
            })
        if len(params) > 0:
            body.append(Opener(params))
        disp = Flex(body, direction='column')
        if self.pane is None:
            self.pane = VariableDisplay('current_pane', value=disp, namespace=self.id)
            return self.pane
        else:
            return disp

    def module_browser(self,
                       members=None,
                       name=None,
                       id=None,
                       details=None,
                       related=None,
                       description=None,
                       examples=None,
                       tests=None,
                       lineno=None,
                       _self=None,
                       **kw
                       ):
        from ..Jupyter import JHTML, Flex, Opener, CardOpener, Button, LinkButton, VariableDisplay, Var

        body = [
            JHTML.SubsubHeading(JHTML.Code("module"), " ",
                                LinkButton(name, lambda *_,**__:subprocess.call(['open', _self.obj.__file__]))
                                ),
            JHTML.Markdown(description),
            CardOpener({
                JHTML.Bold(k.split(".")[-1]):
                    _self.tree.get(k, {
                        # 'output': Button("Click to load...",
                        #                  lambda *_, **__: Var(k, namespace=id).set_value(k)
                        #                  )
                        'output': VariableDisplay(
                            Var(k, namespace=id),
                            value=Button("Click to load...",
                                         lambda *_, k=k, **__: Var(k, namespace=id).set_value(
                                             jdoc(getattr(_self.obj, k.split(".")[-1]), engine=self)
                                         ))
                        )
                    })['output']
                for k in members.keys()
            })
        ]

        # print(
        #     [_self.tree[k] for k in members.keys()]
        # )
        params = self.clean_params({
            # 'ID': id,
            JHTML.SubsubsubHeading('Details'): details,
            JHTML.SubsubsubHeading('Examples'): examples,
            JHTML.SubsubsubHeading('Tests'): tests,
            JHTML.SubsubsubHeading('Related'): related
        })
        if len(params) > 0:
            body.append(Opener(params))

        return Flex(body, direction='column')

    # @classmethod
    # def make_parameter_table(cls, params):
    #     from ..Jupyter import JHTML, Flex, Opener
    #
    #     return Flex([
    #         JHTML.Code()
    #     ])

    def class_browser(self,
                      id=None,
                      name=None,
                      related=None,
                      out_file=None,
                      lineno=None,
                      parameters=None,
                      props=None,
                      description=None,
                      methods=None,
                      examples=None,
                      tests=None,
                      details=None,
                      _self=None,
                      **_
                      ):
        from ..Jupyter import JHTML, Flex, Opener, LinkButton

        body = [
            JHTML.SubsubHeading(JHTML.Code("class"), " ",
                                LinkButton(name, lambda *_, **__: subprocess.call(['open', sys.modules[_self.obj.__module__].__file__]))
                                ),
            JHTML.Markdown(description) if description is not None and len(description) > 0 else ""
        ]
        if len(props) > 0:
            body.append(self.format_props_table(_self, props))
        if len(parameters) > 0:
            body.append(self.format_parameters_table(parameters))
        body.extend(m.handle() for m in methods)

        sections = self.prep_pars(
            _self,
            self.clean_params({
                # 'ID': id,
                'Details': details,
                'Examples': examples,
                'Related': related,
            })
        )
        if len(sections) > 0:
            body.append(Opener(sections))
        return Flex(body, direction='column')

    def method_browser(self,
                       id=None,
                       name=None,
                       decorator=None,
                       signature=None,
                       related=None,
                       out_file=None,
                       lineno=None,
                       parameters=None,
                       props=None,
                       description=None,
                       examples=None,
                       tests=None,
                       details=None,
                       **_
                       ):
        from ..Jupyter import JHTML, Flex, Opener

        header = JHTML.Pre(JHTML.Code(
            # decorator+
            # "def "+name+signature+":",
            JHTML.Span(decorator, cls="cm-meta"),
            JHTML.Span("def", cls="cm-keyword"),
            " ",
            JHTML.Span(name, cls="cm-def"),
            signature + ":",
            cls="cm-s-jupyter language-python"
        ))

        body = Flex(
            [
                JHTML.Markdown(description) if description is not None and len(description.strip()) > 0 else "",
                self.format_parameters_table(parameters),
                Opener({
                    JHTML.SubsubsubHeading(k): JHTML.Markdown(v) for k, v in
                    self.clean_params({
                        # 'ID': id,
                        'Details': details,
                        'Examples': examples
                    }).items()
                })
            ],
            direction='column'
        )

        return Opener({header:body}, theme={
            'header':{
                'wrapper':{'cls':['p-0', 'bg-transparent']}, 'item':{'cls':['pb-0', 'pt-1', 'ps-2']}
            },
            'body':{'wrapper':{'cls':['ps-3']}}
        })

    def object_browser(self,
                       id=None,
                       name=None,
                       related=None,
                       out_file=None,
                       lineno=None,
                       parameters=None,
                       props=None,
                       description=None,
                       methods=None,
                       examples=None,
                       tests=None,
                       details=None,
                       _self=None,
                       **_
                       ):
        from ..Jupyter import JHTML, Flex, Opener

        return Flex(
            [
                JHTML.SubsubHeading(JHTML.Code(type(_self.obj).__name__), " ", name),
                JHTML.Markdown(description),
                str(parameters),
                Opener({
                    JHTML.SubsubsubHeading(k): JHTML.Markdown(v) for k, v in
                    self.clean_params({
                        # 'ID': id,
                        'Details': details,
                        'Examples': examples
                    }).items()
                })
            ],
            direction='column'
        )

    def function_browser(self,
                         id=None,
                         name=None,
                         decorator=None,
                         signature=None,
                         related=None,
                         out_file=None,
                         lineno=None,
                         parameters=None,
                         props=None,
                         description=None,
                         examples=None,
                         tests=None,
                         details=None,
                         **_
                         ):
        from ..Jupyter import JHTML, Flex, Opener

        return Flex(
            [
                JHTML.SubsubHeading(JHTML.Code("function"), " ", name),
                JHTML.Div(
                    JHTML.Pre(JHTML.Code(
                        JHTML.Span("def", cls="cm-keyword"),
                        " ",
                        JHTML.Span(name, cls="cm-def"),
                        signature + ":",
                        cls="cm-s-jupyter language-python"
                    )),
                    JHTML.Div(
                        JHTML.Markdown(description) if description is not None and len(description.strip()) > 0 else "",
                        self.format_parameters_table(parameters),
                        cls='ps-3'
                    ),
                    cls='ps-4'
                ),
                Opener({
                    JHTML.SubsubsubHeading(k): JHTML.Markdown(v) for k, v in
                    self.clean_params({
                        # 'ID': id,
                        'Details': details,
                        'Examples': examples
                    }).items()
                })
            ],
            direction='column'
        )


class DocTemplateHandler(TemplateHandler):
    protected_fields = {'id'}
    default_fields = {'details': "", "related": ""}
    def __init__(self,
                 obj,
                 *,
                 out=None,
                 engine: TemplateEngine=None,
                 root=None,
                 examples_loader:ExamplesExtractor=None,
                 tests_loader:TestsExtractor=None,
                 include_line_numbers=True,
                 walker:'TemplateWalker'=None,
                 **extra_fields
                 ):

        if walker is None:
            walker = DocWalker()
        if engine is None:
            engine = walker.engine

        self.examples_loader = examples_loader
        self.tests_loader = tests_loader
        self.include_line_numbers = include_line_numbers
        super().__init__(obj, out=out, engine=engine, root=root, walker=walker, **extra_fields)

    def get_lineno(self):
        # try:
        if self.include_line_numbers:
            obj = self.obj
            if isinstance(obj, (classmethod, staticmethod, property, types.MethodDescriptorType)):
                try:
                    obj = obj.fget
                except AttributeError:
                    obj = obj.__func__
            elif not isinstance(obj, (type, types.ModuleType, types.FunctionType, types.MethodDescriptorType, types.MethodType)):
                obj = type(obj)
            try:
                lineno = 1 + inspect.findsource(obj)[1] if self.include_line_numbers else ""
            except:
                lineno = ""
        else:
            lineno = ""
        # except:
        #     lineno = ""
        return lineno

    def parse_doc(self, doc):
        """

        :param doc:
        :type doc: str
        :return:
        :rtype:
        """

        # parses a docstring based on reStructured text type specs but Markdown description
        splits = inspect.cleandoc(doc.strip()).splitlines()

        params = collections.deque()
        param_map = {}
        extra_fields = {}

        active_tag = None
        description = []
        for line in splits:
            if line.startswith(":"):
                if line.startswith(":param"):
                    bits = line.split(":", 2)[1:]
                    name = bits[0][5:].strip()
                    active_tag = name
                    if name not in param_map:
                        params.appendleft(name)
                        param_map[name] = {"name": name, "type": "Any", "description": []}
                    desc = bits[1].strip() if len(bits) == 2 else ""
                    if len(desc) > 0:
                        param_map[name]["description"].append(desc)
                elif line.startswith(":type"):
                    bits = line.split(":", 2)[1:]
                    name = bits[0][4:].strip()
                    active_tag = name
                    if name not in param_map:
                        params.appendleft(name)
                        param_map[name] = {"name": name, "type": "Any", "description": []}
                    t = bits[1].strip() if len(bits) == 2 else ""
                    if len(t) > 0:
                        param_map[name]["type"] = t
                elif line.startswith(":return"):
                    bits = line.split(":", 2)[1:]
                    name = ":returns"
                    active_tag = name
                    if name not in param_map:
                        params.appendleft(name)
                        param_map[name] = {"name": name, "type": "_", "description": []}
                    t = bits[1].strip() if len(bits) == 2 else ""
                    if len(t) > 0:
                        param_map[name]["description"] = t
                elif line.startswith(":rtype"):
                    bits = line.split(":", 2)[1:]
                    name = ":returns"
                    active_tag = name
                    if name not in param_map:
                        params.appendleft(name)
                        param_map[name] = {"name": name, "type": "_", "description": []}
                    t = bits[1].strip() if len(bits) == 2 else ""
                    if len(t) > 0:
                        param_map[name]["type"] = t
                else:
                    split = line.split()[0]
                    if len(split) > 2 and split.endswith(":"):
                        bits = line.split(":", 2)[1:]
                        name = bits[0]
                        active_tag = name
                        t = bits[1].strip() if len(bits) == 2 else ""
                        if len(t) > 0:
                            extra_fields[name] = [t]
                        else:
                            extra_fields[name] = []
            else:
                if active_tag is None:
                    description.append(line.strip())
                else:
                    if active_tag in param_map:
                        param_map[active_tag]['description'].append(line)
                    else:
                        extra_fields[active_tag].append(line)
        for v in param_map.values():
            v['description'] = "\n".join(v['description'])
        for k, v in extra_fields.items():
            extra_fields[k] = "\n".join(v)

        return inspect.cleandoc("\n".join(description)).strip(), param_map, extra_fields

    def load_examples(self):
        return self.examples_loader.load(self) if self.examples_loader is not None else None
    def load_tests(self):
        tests = self.tests_loader.load(self) if self.tests_loader is not None else None
        if tests is None:
            parent_tests = self.resolve_key('parent_tests')
            if parent_tests is not None:
                tests = parent_tests.filter_by_name(self.name.split(".")[-1])
        if tests is not None:
            self.extra_fields['parent_tests'] = tests
            tests = TestExamplesFormatter(tests).get_template_parameters()
        return tests

class DocObjectTemplateHandler(DocTemplateHandler):
    def get_package_and_url(self, include_url_base=True):
        pkg, file_url=super().get_package_and_url(include_url_base=include_url_base)
        try:
            base_url, ext = file_url.rsplit("/", 1)
        except ValueError:
            pass
        else:
            if ext == '__init__.py':
                try:
                    base_url, ext = base_url.rsplit("/", 1)
                except ValueError:
                    pass
            file_url = base_url + '.py'
        return pkg, file_url

    def load_examples(self):
        return self.examples_loader.load(self) if self.examples_loader is not None else None
    def load_tests(self):
        tests = self.tests_loader.load(self) if self.tests_loader is not None else None
        if tests is None:
            parent_tests = self.resolve_key('parent_tests')
            if parent_tests is not None:
                tests = parent_tests.filter_by_name(self.name.split(".")[-1])
        if tests is not None:
            self.extra_fields['parent_tests'] = tests
            tests = TestExamplesFormatter(tests).get_template_parameters()
        return tests
class ModuleWriter(DocTemplateHandler):
    """
    A writer targeted to a module object. Just needs to write the Module metadata.

    :related: DocWalker
    """
    template = 'module.md'
    def __init__(self, obj, **kwargs):
        if isinstance(obj, str):
            obj = importlib.import_module(obj)
        super().__init__(obj, **kwargs)

    def get_template_params(self):
        """
        Provides module specific parameters
        :return:
        :rtype:
        """

        mod = self.obj  # type: types.ModuleType
        name = mod.__name__
        ident = self.get_target_extension(self.identifier)
        ident_depth = len(ident)
        ident = ".".join(ident)
        # get identifiers
        mems = [inspect.getattr_static(mod, a) for a in self.get_members(mod)]
        idents = {
            self.get_identifier(a):".".join(self.get_target_extension(a))
            for a in mems
        }
        # flattend them
        idents = {a:i for a,i in idents.items() if ident in i}
        # split by qualified names
        idents = {a:".".join(a.split(".")[ident_depth-1:]) for a in idents}
        descr, _, fields = self.parse_doc(mod.__doc__ if mod.__doc__ is not None else '')
        # long_descr = mod.__long_doc__ if hasattr(mod, '__long_doc__') and mod.__long_doc__ is not None else ''

        ex = self.load_examples()
        tests = self.load_tests()
        return dict({
            'id': ident,
            'description': descr.strip(),
            # 'long_description': long_descr.strip(),
            'name': name,
            'members': idents,
            'examples': ex,
            'tests': tests,
            'lineno' : self.get_lineno()
        }, **fields)

    @classmethod
    def get_members(cls, mod):
        return (mod.__all__ if hasattr(mod, '__all__') else [])

class ClassWriter(DocObjectTemplateHandler):
    """
    A writer targeted to a class

    :related: DocWalker
    """

    template = 'class.md'
    def load_methods(self, function_writer=None):
        """
        Loads the methods supported by the class

        :param function_writer:
        :type function_writer:
        :return:
        :rtype:
        """

        if function_writer is None:
            function_writer = self.walker.method_handler

        cls = self.obj
        keys = cls.__all__ if hasattr(cls, '__all__') else list(cls.__dict__.keys())

        props = []
        methods = []

        extra_fields = self.extra_fields.copy()
        pkg, file_url = self.package_path
        extra_fields['package_name'] = pkg
        extra_fields['file_url'] = file_url
        extra_fields['package_url'] = os.path.dirname(file_url)
        for k in keys:
            o = inspect.getattr_static(cls, k)
            if isinstance(o, (types.FunctionType, types.MethodType, classmethod, staticmethod, property)):
                if not k.startswith("_") or (k.startswith("__") and k.endswith("__")):
                    methods.append(
                        self.walker.get_handler(
                            o,
                            cls=function_writer,
                            name=k,
                            out_file=None,
                            extra_fields=extra_fields,
                            parent=self.identifier
                        )
                    )
            else:
                if not k.startswith("_"):
                    props.append((k, o))

        return props, methods

    def format_prop(self, k, o):
        return '{}: {}'.format(k, type(o).__name__)

    def get_template_params(self, function_writer=None):
        """

        :param function_writer:
        :type function_writer:
        :return:
        :rtype:
        """

        cls = self.obj  # type: type
        ex = self.load_examples()
        tests = self.load_tests()
        name = cls.__name__
        ident = self.identifier
        props, methods = self.load_methods(function_writer=function_writer)
        descr, param, fields = self.parse_doc(cls.__doc__ if cls.__doc__ is not None else '')
        lineno = self.get_lineno()

        return dict({
            'id': ident,
            'name': name,
            'lineno': lineno,
            'description': descr,
            'parameters': param,
            'props': props,
            'methods': methods,
            'examples': ex if ex is not None else "",
            'tests': tests
        }, **fields)

class FunctionWriter(DocObjectTemplateHandler):
    """
    Writer to dump functions to file

    :related: DocWalker

    """

    template = 'function.md'

    def get_signature(self):
        return str(inspect.signature(self.obj))

    def get_template_params(self, **kwargs):
        f = self.obj  # type: types.FunctionType
        ident = self.identifier
        signature = self.get_signature()
        mem_obj_pat = re.compile(" object at \w+>")
        signature = re.sub(mem_obj_pat, " instance>", signature)
        name = self.get_name()
        descr, param, fields = self.parse_doc(f.__doc__ if f.__doc__ is not None else '')
        ex = self.load_examples()
        tests = self.load_tests()
        lineno = self.get_lineno()
        return dict({
            'id': ident,
            'name': name,
            'lineno': lineno,
            'signature': signature,
            'parameters': param,
            'description': descr,
            'examples': ex if ex is not None else "",
            'tests': tests
        }, **fields)

class MethodWriter(FunctionWriter):
    """
    Writes class methods to file
    (distinct from functions since not expected to exist solo)
    """

    template = 'method.md'

    def get_template_params(self, **kwargs):
        meth = self.obj
        try:
            if isinstance(meth, (classmethod, staticmethod, property, types.MethodDescriptorType)):
                try:
                    self.obj = meth.fget
                except AttributeError:
                    self.obj = meth.__func__
            params = super().get_template_params(**kwargs)
        finally:
            self.obj = meth
        decorator = ""
        if isinstance(meth, classmethod):
            decorator = 'classmethod'
        elif isinstance(meth, property):
            decorator = 'property'
        elif isinstance(meth, staticmethod):
            decorator = 'staticmethod'
        if len(decorator) > 0:
            decorator = "@" + decorator + "\n"
        params['decorator'] = decorator
        return params

    def get_signature(self):
        try:
            signature = str(inspect.signature(self.obj))
        except TypeError:  # dies on properties
            signature = "(self)"
        return signature

    @property
    def identifier(self):
        if isinstance(self.obj, property):
            return self.get_identifier(self.resolve_parent(check_tree=False)) + "." + self.get_name()
        else:
            return self.get_identifier(self.obj)

class ObjectWriter(DocObjectTemplateHandler):
    """
    Writes general objects to file.
    Basically a fallback to support singletons and things
    of that nature.

    :related: DocWalker

    """

    template = 'object.md'

    @property
    def identifier(self):
        try:
            qualname = self.obj.__qualname__
        except AttributeError:
            qualname = self.get_identifier(type(self.obj)) + "." + self.get_name()
        qn = qualname.split(".")
        qualname = ".".join(qn[:-2] + qn[-1:])  # want to drop the class name
        return qualname

    def check_should_write(self):
        """
        Determines whether the object really actually should be
        documented (quite permissive)
        :return:
        :rtype:
        """
        return (
                hasattr(self.obj, "__doc__")
                and hasattr(self.obj, "__name__")
                and super().check_should_write()
        )

    def get_template_params(self):

        try:
            doc = self.obj.__doc__
        except AttributeError:
            descr = "instance of " + type(self.obj).__name__
            fields = {}
        else:
            descr, _, fields = self.parse_doc(doc if doc is not None else '')

        if descr is None:
            descr = ''

        ex = self.load_examples()
        lineno = self.get_lineno()
        return dict({
            'id': self.identifier,
            'lineno': lineno,
            'name': self.get_name(),
            'description': descr.strip(),
            'examples': ex if ex is not None else "",
            'tests': None
        }, **fields)

class IndexWriter(DocTemplateHandler):
    """
    Writes an index file with all of the
    written documentation files.
    Needs some work to provide more useful info by default.

    :related: DocWalker

    """
    template = 'index.md'

    def __init__(self, *args, description=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.description = description if description is not None else "# Documentation"

    def get_identifier(cls, o):
        return 'index'

    def get_file_paths(self):
        rl = len(os.path.split(self.root)) if self.root is not None else 0
        fs = ["/".join(os.path.split(f)[rl - 1:]) if isinstance(f, str) else f for f in self.obj]
        return fs
    def get_index_files(self):
        return [
            [os.path.splitext(os.path.basename(f))[0], f] if isinstance(f, str) else f
            for f in self.get_file_paths()
        ]

    # def format_item(self, f):
    #     self.formatter.format_item(
    #         self.formatter.format_link(os.path.splitext(f.split("/")[-1])[0], f)
    #     )

    def get_template_params(self):
        descr, _, fields = self.parse_doc(self.description if self.description is not None else '')
        return dict({
            'index_files': self.get_index_files(),
            'description': descr,
            'examples': self.load_examples()
        }, **fields)

class DocWalker(TemplateWalker):
    """
    A class that walks a module structure, generating `.md` files for every class inside it as well as for global functions,
    and a Markdown index file.

    Takes a set of objects & writers and walks through the objects, generating files on the way.

    :details: A `DocWalker` object is a light subclass of a `TemplateWalker`, but specialized for documentation & with specialized handlers
    :related: .DocsBuilder.DocBuilder, ModuleWriter, ClassWriter, FunctionWriter, MethodWriter, ObjectWriter, IndexWriter
    """

    module_handler = ModuleWriter
    class_handler = ClassWriter
    function_handler = FunctionWriter
    method_handler = MethodWriter
    object_handler = ObjectWriter
    index_handler = IndexWriter
    spec = DocSpec
    def __init__(self,
                 out=None,
                 engine=None,
                 verbose=True,
                 template_locator=None,
                 examples_directory=None,
                 tests_directory=None,
                 **extra_fields
                 ):
        """
        :param objects: the objects to write out
        :type objects: Iterable[Any]
        :param out: the directory in which to write the files (`None` means `sys.stdout`)
        :type out: None | str
        :param ignore_paths: a set of paths not to write (passed to the objects)
        :type ignore_paths: None | Iterable[str]
        """

        if engine is None:
            engine = self.get_engine(template_locator)
        self.examples_loader = self.get_examples_loader(examples_directory)
        self.tests_loader = self.get_tests_loader(tests_directory)
        self.verbose = verbose

        super().__init__(engine, out=out, **extra_fields)

    def get_engine(self, locator):
        if locator is None:
            locator = InteractiveTemplateEngine()
        elif not isinstance(locator, TemplateEngine):
            locator = TemplateEngine(locator, template_pattern="*.md", formatter_class=MarkdownTemplateFormatter)
        return locator
    def get_examples_loader(self, examples_directory):
        examples_loader = examples_directory
        if examples_loader is not None and not isinstance(examples_loader, ExamplesExtractor):
            examples_loader = ExamplesExtractor(examples_loader)
        return examples_loader
    def get_tests_loader(self, tests_directory):
        tests_loader = tests_directory
        if tests_loader is not None and not isinstance(tests_loader, TestsExtractor):
            tests_loader = TestsExtractor(tests_loader)
        return tests_loader
    def get_handler(self, *args, examples_loader=None, tests_loader=None, **kwargs):
        return super().get_handler(
            *args,
            examples_loader=self.examples_loader if examples_loader is None else examples_loader,
            tests_loader=self.tests_loader if tests_loader is None else tests_loader,
            **kwargs
        )

    def visit_root(self, o, tests_directory=None, examples_directory=None, **kwargs):
        if tests_directory is None and isinstance(o, dict):
            tests_directory = o.get('tests_directory', None)
        old_tl = self.tests_loader
        if tests_directory is not None:
            self.tests_loader = self.get_tests_loader(tests_directory)
        if examples_directory is None and isinstance(o, dict):
            examples_directory = o.get('examples_directory', None)
        old_el = self.examples_loader
        if examples_directory is not None:
            self.examples_loader = self.get_examples_loader(examples_directory)
        try:
            return super().visit_root(o, **kwargs)
        finally:
            self.tests_loader = old_tl
            self.examples_loader = old_el

    @mixedmethod
    def _ipython_pinfo_(cls):
        return jdoc(cls)


def jdoc(obj, max_depth=1, engine=None):
    """
    provides documentation in a Jupyter-friendly environment

    :param obj: the object to extract documentation for
    :type obj: Any
    :param max_depth: the depth in the object tree to go down to (default: `1`)
    :type max_depth: int
    :return docs:
    :rtype docs: ..Jupyter.Component

    :details:
    Makes use of the `JHTML` system to nicely format documentation as well as the
    documentation utilities found in `McUtils.Docs` (which were orginally written
    as part of the `Peeves` package).

    Asking for too many pieces of documentation at once can really slow things down,
    so by default the object tree is traversed only shallowly.
    In principle, the documentation could be directly exported to an HTML document as
    no `Widget`-necessary bits are used (at least as on when I write this) but a better
    choice is to generate Markdown docs using the standard `DocsBuilder` approach.
    This provides flexibility and can be ingested into any number of Markdown->HTML systems.

    :examples:
    Get documentation for `jdoc`

    ```python
    jdoc(jdoc)
    ```

    Get documentation for an entire module

    ```python
    import McUtils.Docs as MDoc
    jdoc(Mdoc)
    ```

    """
    return DocWalker(
        description="",
        engine=engine
    ).write([obj], max_depth=max_depth)